import pytest
from ppp_connectors.api_connectors.broker import Broker

def test_get_calls_make_request(mocker):
    broker = Broker(base_url="https://example.com")
    mock_request = mocker.patch.object(broker, "_make_request")
    broker.get("/test", params={"foo": "bar"})
    mock_request.assert_called_once_with("GET", "/test", params={"foo": "bar"})

def test_post_calls_make_request(mocker):
    broker = Broker(base_url="https://example.com")
    mock_request = mocker.patch.object(broker, "_make_request")
    broker.post("/submit", json={"data": 123})
    mock_request.assert_called_once_with("POST", "/submit", json={"data": 123})

def test_logging_enabled_logs_message(mocker):
    mock_logger = mocker.MagicMock()
    mock_setup_logger = mocker.patch("ppp_connectors.api_connectors.broker.setup_logger", return_value=mock_logger)
    broker = Broker(base_url="https://example.com", enable_logging=True)
    broker._log("test message")
    mock_logger.info.assert_called_once_with("test message")

def test_env_only_proxy_from_dotenv(monkeypatch, mocker):
    # Simulate .env via env_config when load_env_vars=True
    mock_combine = mocker.patch(
        "ppp_connectors.api_connectors.broker.combine_env_configs",
        return_value={"HTTPS_PROXY": "http://env-proxy:8080"}
    )
    broker = Broker(base_url="https://example.com", load_env_vars=True, trust_env=False)
    assert broker.proxy == "http://env-proxy:8080"
    assert broker.mounts is None
    mock_combine.assert_called_once()

def test_env_only_proxy_from_osenv(monkeypatch):
    # Simulate OS env with trust_env=True
    monkeypatch.setenv("HTTPS_PROXY", "http://os-proxy:9090")
    broker = Broker(base_url="https://example.com", load_env_vars=False, trust_env=True)
    assert broker.proxy == "http://os-proxy:9090"
    assert broker.mounts is None

def test_env_per_scheme_mounts(monkeypatch):
    # Different HTTP and HTTPS proxies from env
    monkeypatch.setenv("HTTP_PROXY", "http://http-proxy:8000")
    monkeypatch.setenv("HTTPS_PROXY", "http://https-proxy:9000")
    broker = Broker(base_url="https://example.com", load_env_vars=False, trust_env=True)
    assert broker.proxy is None
    assert "http://" in broker.mounts and "https://" in broker.mounts

def test_priority_mounts_over_proxy(monkeypatch):
    # If mounts are provided, they win
    mounts = {"http://": object(), "https://": object()}
    broker = Broker(base_url="https://example.com", mounts=mounts, proxy="http://should-not-use:1111")
    assert broker.mounts == mounts
    assert broker.proxy == "http://should-not-use:1111"  # Stored but not used for session if mounts present

def test_priority_proxy_over_env(monkeypatch):
    monkeypatch.setenv("HTTPS_PROXY", "http://os-proxy:9999")
    broker = Broker(base_url="https://example.com", proxy="http://explicit-proxy:7777", trust_env=True)
    assert broker.proxy == "http://explicit-proxy:7777"

def test_no_proxy_when_nothing_set(monkeypatch):
    monkeypatch.delenv("ALL_PROXY", raising=False)
    monkeypatch.delenv("HTTP_PROXY", raising=False)
    monkeypatch.delenv("HTTPS_PROXY", raising=False)
    broker = Broker(base_url="https://example.com", load_env_vars=False, trust_env=False)
    assert broker.proxy is None
    assert broker.mounts is None



import httpx
from httpx import Auth
from typing import Optional, Dict, Any, Union, Iterable, Callable, ParamSpec, TypeVar
from tenacity import retry, stop_after_attempt, wait_exponential, RetryError, retry_if_exception, AsyncRetrying
from ppp_connectors.helpers import setup_logger, combine_env_configs
from functools import wraps
import inspect
import os


P = ParamSpec("P")
R = TypeVar("R")

def log_method_call(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        caller = func.__name__
        sig = inspect.signature(func)
        bound = sig.bind(self, *args, **kwargs)
        bound.apply_defaults()
        query_value = bound.arguments.get("query")
        self._log(f"{caller} called with query: {query_value}")
        return func(self, *args, **kwargs)
    return wrapper


def bubble_broker_init_signature(*, exclude: Iterable[str] = ("base_url",)):
    """
    Class decorator that augments a connector subclass' __init__ signature with
    parameters from Broker.__init__ for better IDE/tab-completion hints.

    Usage:
        from ppp_connectors.api_connectors.broker import Broker, bubble_broker_init_signature

        @bubble_broker_init_signature()
        class MyConnector(Broker):
            def __init__(self, api_key: str | None = None, **kwargs):
                super().__init__(base_url="https://example.com", **kwargs)
                ...

    Notes:
        - This affects *introspection only* (via __signature__). Runtime behavior is unchanged.
        - Subclass-specific parameters remain first (e.g., api_key), followed by Broker params.
        - `base_url` is excluded by default since subclasses set it themselves.
        - The subclass' **kwargs (if present) is preserved at the end so httpx.Client kwargs
          can still be passed through.
    """
    def _decorate(cls):
        sub_init = cls.__init__
        broker_init = Broker.__init__

        sub_sig = inspect.signature(sub_init)
        broker_sig = inspect.signature(broker_init)

        new_params = []
        saw_var_kw = None

        # Keep subclass params first; remember its **kwargs if present
        for p in sub_sig.parameters.values():
            if p.kind is inspect.Parameter.VAR_KEYWORD:
                saw_var_kw = p
            else:
                new_params.append(p)

        present = {p.name for p in new_params}

        # Append Broker params (skip self, excluded, already-present, and **kwargs)
        for name, p in list(broker_sig.parameters.items())[1:]:
            if name in exclude or name in present:
                continue
            if p.kind is inspect.Parameter.VAR_KEYWORD:
                continue
            new_params.append(p)

        # Re-append subclass **kwargs (or add a generic one to keep flexibility)
        if saw_var_kw is not None:
            new_params.append(saw_var_kw)
        else:
            new_params.append(
                inspect.Parameter(
                    "client_kwargs",
                    kind=inspect.Parameter.VAR_KEYWORD,
                )
            )

        cls.__init__.__signature__ = inspect.Signature(parameters=new_params)
        return cls

    return _decorate


class SharedConnectorBase:
    """
    Shared base class for Broker and AsyncBroker.
    Houses reusable logic (constructor, logging, proxy config, retry predicate).
    """
    def __init__(
        self,
        base_url: str,
        headers: Optional[Dict[str, str]] = None,
        enable_logging: bool = False,
        enable_backoff: bool = False,
        timeout: int = 10,
        load_env_vars: bool = False,
        trust_env: bool = True,
        proxy: Optional[str] = None,
        mounts: Optional[Dict[str, httpx.HTTPTransport]] = None,
        **client_kwargs,
    ):
        self.base_url = base_url.rstrip('/')
        self.logger = setup_logger(self.__class__.__name__) if enable_logging else None
        self.enable_backoff = enable_backoff
        self.timeout = timeout
        self.headers = headers or {}
        self.trust_env = trust_env
        self.proxy = proxy
        self.mounts = mounts
        self.env_config = combine_env_configs() if load_env_vars else {}
        self._client_kwargs = dict(client_kwargs) if client_kwargs else {}

    def _log(self, message: str):
        if self.logger:
            self.logger.info(message)

    def _collect_proxy_config(self) -> tuple[Optional[str], Optional[Dict[str, httpx.HTTPTransport]]]:
        source_env: Optional[Dict[str, str]] = None
        if isinstance(self.env_config, dict) and len(self.env_config) > 0:
            source_env = {k: v for k, v in self.env_config.items() if isinstance(k, str) and isinstance(v, str)}
        elif self.trust_env:
            source_env = dict(os.environ)
        else:
            return None, None

        def _get(key: str) -> Optional[str]:
            return source_env.get(key) or source_env.get(key.lower())

        all_proxy = _get("ALL_PROXY")
        http_proxy = _get("HTTP_PROXY")
        https_proxy = _get("HTTPS_PROXY")

        if http_proxy and https_proxy and http_proxy != https_proxy:
            return None, {
                "http://": httpx.HTTPTransport(proxy=http_proxy),
                "https://": httpx.HTTPTransport(proxy=https_proxy),
            }
        single = all_proxy or https_proxy or http_proxy
        if single:
            return single, None
        return None, None

    @staticmethod
    def _default_retry_exc(exc: BaseException) -> bool:
        if isinstance(exc, httpx.HTTPStatusError):
            r = exc.response
            if r is not None:
                return r.status_code == 429 or 500 <= r.status_code < 600
        return isinstance(exc, (
            httpx.ConnectError,
            httpx.ReadTimeout,
            httpx.WriteError,
            httpx.RemoteProtocolError,
            httpx.PoolTimeout,
        ))


class Broker(SharedConnectorBase):
    """
    A base HTTP client that provides structured request handling, logging, retries, and optional environment config loading.
    Designed to be inherited by specific API connector classes.
    """
    def __init__(
        self,
        base_url: str,
        headers: Optional[Dict[str, str]] = None,
        enable_logging: bool = False,
        enable_backoff: bool = False,
        timeout: int = 10,
        load_env_vars: bool = False,
        trust_env: bool = True,
        proxy: Optional[str] = None,
        mounts: Optional[Dict[str, httpx.HTTPTransport]] = None,
        **client_kwargs,
    ):
        super().__init__(
            base_url=base_url,
            headers=headers,
            enable_logging=enable_logging,
            enable_backoff=enable_backoff,
            timeout=timeout,
            load_env_vars=load_env_vars,
            trust_env=trust_env,
            proxy=proxy,
            mounts=mounts,
            **client_kwargs,
        )

        client_options = dict(self._client_kwargs)
        client_options.pop("timeout", None)

        client_args = {
            "timeout": self.timeout,
            "trust_env": self.trust_env,
            **client_options,
        }

        if self.mounts:
            client_args["mounts"] = self.mounts
        elif self.proxy:
            client_args["proxy"] = self.proxy
        else:
            env_proxy, env_mounts = self._collect_proxy_config()
            if env_mounts:
                self.mounts = env_mounts
                client_args["mounts"] = self.mounts
            elif env_proxy:
                self.proxy = env_proxy
                client_args["proxy"] = self.proxy
            elif not self.trust_env:
                client_args["trust_env"] = False

        self.session = httpx.Client(**client_args)

    def _make_request(
        self,
        method: str,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        auth: Optional[Union[tuple, Auth]] = None,
        headers: Optional[Dict[str, str]] = None,
        retry_kwargs: Optional[Dict[str, Any]] = None,
        **request_kwargs,
    ) -> httpx.Response:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"

        def do_request() -> httpx.Response:
            resp = self.session.request(
                method=method,
                url=url,
                headers=headers or self.headers,
                params=params,
                json=json,
                auth=auth,
                **request_kwargs,
            )
            resp.raise_for_status()
            return resp

        call = do_request
        if self.enable_backoff:
            rk = dict(retry_kwargs or {})
            if "retry" not in rk:
                rk["retry"] = retry_if_exception(self._default_retry_exc)
            if "stop" not in rk:
                rk["stop"] = stop_after_attempt(3)
            if "wait" not in rk:
                rk["wait"] = wait_exponential(multiplier=1, min=2, max=10)
            call = retry(reraise=True, **rk)(do_request)

        try:
            return call()
        except RetryError as re:
            last = re.last_attempt.exception()
            self._log(f"Retry failed: {last}")
            raise
        except httpx.HTTPStatusError as he:
            self._log(f"HTTP error: {he}")
            raise

    def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None, **kwargs) -> httpx.Response:
        return self._make_request("GET", endpoint, params=params, **kwargs)

    def post(self, endpoint: str, json: Optional[Dict[str, Any]] = None, **kwargs) -> httpx.Response:
        return self._make_request("POST", endpoint, json=json, **kwargs)


class AsyncBroker(SharedConnectorBase):
    """
    Async HTTP client connector. Provides async _make_request, get, and post.
    """
    def __init__(
        self,
        base_url: str,
        headers: Optional[Dict[str, str]] = None,
        enable_logging: bool = False,
        enable_backoff: bool = False,
        timeout: int = 10,
        load_env_vars: bool = False,
        trust_env: bool = True,
        proxy: Optional[str] = None,
        mounts: Optional[Dict[str, httpx.HTTPTransport]] = None,
        **client_kwargs,
    ):
        super().__init__(
            base_url=base_url,
            headers=headers,
            enable_logging=enable_logging,
            enable_backoff=enable_backoff,
            timeout=timeout,
            load_env_vars=load_env_vars,
            trust_env=trust_env,
            proxy=proxy,
            mounts=mounts,
            **client_kwargs,
        )

        if self.mounts:
            raise ValueError("The 'mounts' parameter is not supported in AsyncBroker but "
                             "you can still use 'proxy' or 'trust_env' if 'HTTP_PROXY' or "
                             "'HTTPS_PROXY' are in your system environment variables ")

        resolved_proxy = self.proxy or self._collect_proxy_config()[0]

        self.session = httpx.AsyncClient(
            timeout=self.timeout,
            proxy=resolved_proxy,
            trust_env=self.trust_env,
            **self._client_kwargs,
        )

    async def _make_request(
        self,
        method: str,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        auth: Optional[Union[tuple, Auth]] = None,
        headers: Optional[Dict[str, str]] = None,
        retry_kwargs: Optional[Dict[str, Any]] = None,
        **request_kwargs,
    ) -> httpx.Response:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"

        async def do_request() -> httpx.Response:
            resp = await self.session.request(
                method=method,
                url=url,
                headers=headers or self.headers,
                params=params,
                json=json,
                auth=auth,
                **request_kwargs,
            )
            resp.raise_for_status()
            return resp

        call = do_request
        if self.enable_backoff:

            rk = dict(retry_kwargs or {})
            retry_pred = rk.get("retry", retry_if_exception(self._default_retry_exc))
            stop_cond = rk.get("stop", stop_after_attempt(3))
            wait_cond = rk.get("wait", wait_exponential(multiplier=1, min=2, max=10))

            async def retry_wrapper():
                async for attempt in AsyncRetrying(reraise=True, retry=retry_pred, stop=stop_cond, wait=wait_cond):
                    with attempt:
                        return await do_request()
            call = retry_wrapper

        try:
            return await call()
        except RetryError as re:
            last = re.last_attempt.exception()
            self._log(f"Retry failed: {last}")
            raise
        except httpx.HTTPStatusError as he:
            self._log(f"HTTP error: {he}")
            raise

    async def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None, **kwargs) -> httpx.Response:
        return await self._make_request("GET", endpoint, params=params, **kwargs)

    async def post(self, endpoint: str, json: Optional[Dict[str, Any]] = None, **kwargs) -> httpx.Response:
        return await self._make_request("POST", endpoint, json=json, **kwargs)


from datetime import date, datetime
from typing import Dict, Any, List, Set, Union, Optional
from httpx import BasicAuth, Response
from ppp_connectors.api_connectors.broker import AsyncBroker, Broker, bubble_broker_init_signature, log_method_call
from ppp_connectors.helpers import validate_date_string

@bubble_broker_init_signature()
class TwilioConnector(Broker):
    """
    Connector for interacting with Twilio Lookup and Usage APIs.

    Supports phone number lookups with data packages and generating account usage reports.
    """

    def __init__(
        self,
        api_sid: Optional[str] = None,
        api_secret: Optional[str] = None,
        **kwargs
    ):
        super().__init__(base_url="https://lookups.twilio.com/v2", **kwargs)

        self.api_sid = api_sid or self.env_config.get("TWILIO_API_SID")
        self.api_secret = api_secret or self.env_config.get("TWILIO_API_SECRET")

        if not self.api_sid or not self.api_secret:
            raise ValueError("TWILIO_API_SID and TWILIO_API_SECRET are required.")

        self.auth = BasicAuth(self.api_sid, self.api_secret)

    @log_method_call
    def lookup_phone(self, phone_number: str, data_packages: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Query information about a phone number using Twilio's Lookup API.

        Args:
            phone_number (str): The phone number to query.
            data_packages (list): Optional data packages to include (e.g. 'caller_name', 'sim_swap').

        Returns:
            httpx.Response: the httpx.Response object
        """
        valid_data_packages: Set[str] = {
            'caller_name', 'sim_swap', 'call_forwarding', 'line_status',
            'line_type_intelligence', 'identity_match', 'reassigned_number',
            'sms_pumping_risk', 'phone_number_quality_score', 'pre_fill'
        }

        if data_packages:
            invalid = set(data_packages) - valid_data_packages
            if invalid:
                raise ValueError(f"Invalid data packages: {', '.join(invalid)}")

        params: Dict[str, Any] = {
            'Fields': ','.join(data_packages) if data_packages else "",
            **kwargs
        }

        endpoint = f"/PhoneNumbers/{phone_number}"
        return self._make_request("get", endpoint=endpoint, auth=self.auth, params=params)


@bubble_broker_init_signature()
class AsyncTwilioConnector(AsyncBroker):
    """
    Async connector for interacting with Twilio Lookup API.
    """

    def __init__(
        self,
        api_sid: Optional[str] = None,
        api_secret: Optional[str] = None,
        **kwargs
    ):
        super().__init__(base_url="https://lookups.twilio.com/v2", **kwargs)

        self.api_sid = api_sid or self.env_config.get("TWILIO_API_SID")
        self.api_secret = api_secret or self.env_config.get("TWILIO_API_SECRET")

        if not self.api_sid or not self.api_secret:
            raise ValueError("TWILIO_API_SID and TWILIO_API_SECRET are required.")

        self.auth = BasicAuth(self.api_sid, self.api_secret)

    @log_method_call
    async def lookup_phone(self, phone_number: str, data_packages: Optional[List[str]] = None, **kwargs) -> Response:
        """
        Async version of phone number lookup using Twilio Lookup API.

        Args:
            phone_number (str): The phone number to query.
            data_packages (list): Optional data packages to include (e.g. 'caller_name', 'sim_swap').

        Returns:
            httpx.Response: the httpx.Response object
        """
        valid_data_packages: Set[str] = {
            'caller_name', 'sim_swap', 'call_forwarding', 'line_status',
            'line_type_intelligence', 'identity_match', 'reassigned_number',
            'sms_pumping_risk', 'phone_number_quality_score', 'pre_fill'
        }

        if data_packages:
            invalid = set(data_packages) - valid_data_packages
            if invalid:
                raise ValueError(f"Invalid data packages: {', '.join(invalid)}")

        params: Dict[str, Any] = {
            'Fields': ','.join(data_packages) if data_packages else "",
            **kwargs
        }

        endpoint = f"/PhoneNumbers/{phone_number}"
        return await self._make_request("get", endpoint=endpoint, auth=self.auth, params=params)


from elasticsearch import Elasticsearch, helpers
from typing import List, Dict, Generator, Any, Optional, Union


try:
    from ppp_connectors.helpers import setup_logger
    _default_logger = setup_logger(name="elasticsearch")
except ImportError:
    import logging
    _default_logger = logging.getLogger("elasticsearch")
    if not _default_logger.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        _default_logger.addHandler(handler)
    _default_logger.setLevel(logging.INFO)


class ElasticsearchConnector:
    """
    A connector class for interacting with Elasticsearch.

    This class provides methods to perform paginated search queries using the scroll API
    and to execute bulk insert operations. It includes integrated logging support for observability.
    """
    def __init__(
        self,
        hosts: List[str],
        username: Optional[str] = None,
        password: Optional[str] = None,
        logger: Optional[Any] = None
    ):
        """
        Initialize the Elasticsearch client.

        Args:
            hosts (List[str]): List of Elasticsearch host URLs.
            username (Optional[str]): Username for basic authentication. Defaults to None.
            password (Optional[str]): Password for basic authentication. Defaults to None.
            logger (Optional[Any]): Optional logger instance. If not provided, a default logger is used.
        """
        self.client = Elasticsearch(hosts, basic_auth=(username, password))
        self.logger = logger if logger is not None else _default_logger

    def _log(self, msg: str, level: str = "info"):
        """
        Internal helper to log messages using the provided or default logger.

        Args:
            msg (str): The message to log.
            level (str): The logging level as a string (e.g., 'info', 'error'). Defaults to 'info'.
        """
        if self.logger:
            log_method = getattr(self.logger, level, self.logger.info)
            log_method(msg)

    def query(
        self,
        index: str,
        query: Union[str, Dict],
        size: int = 1000
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Execute a paginated search query using the Elasticsearch scroll API.

        This method handles retrieval of large result sets by paging through results
        using a scroll context.

        Args:
            index (str): The name of the index to search.
            query (Union[str, Dict]): A Lucene query string or Elasticsearch DSL query body.
            size (int): Number of results to retrieve per batch. Defaults to 1000.

        Yields:
            Generator[Dict[str, Any], None, None]: A generator that yields each search hit as a dictionary.

        Note:
            This method returns a generator. If you want to collect all results,
            you can wrap the result in `list()`, but beware of memory usage if the
            result set is large. Prefer streaming and processing results incrementally.
        """

        self._log(f"Executing query on index '{index}' with batch size {size}", "info")

        if isinstance(query, str):
            query = {
                "query": {
                    "query_string": {
                        "query": query
                    }
                }
            }

        page = self.client.search(index=index, body=query, scroll="2m", size=size)
        sid = page["_scroll_id"]
        hits = page["hits"]["hits"]
        yield from hits

        while hits:
            page = self.client.scroll(scroll_id=sid, scroll="2m")
            sid = page["_scroll_id"]
            hits = page["hits"]["hits"]
            if not hits:
                break
            yield from hits
        self.client.clear_scroll(scroll_id=sid)
        self._log(f"Completed scrolling query on index '{index}'", "info")

    def bulk_insert(
        self,
        index: str,
        data: List[Dict],
        id_key: str = "_id"
    ):
        """
        Perform a bulk insert operation into the specified Elasticsearch index.

        This method sends batches of documents for indexing in a single API call.
        Each document can optionally specify an ID via the `id_key`.

        Args:
            index (str): The name of the index to insert documents into.
            data (List[Dict]): A list of documents to insert.
            id_key (str): The key in each document to use as the document ID. Defaults to "_id".

        Returns:
            Tuple[int, List[Dict]]: A tuple containing the number of successfully processed actions
                                    and a list of any errors encountered during insertion.
        """
        self._log(f"Inserting {len(data)} documents into index '{index}'", "info")
        actions = [
            {
                "_index": index,
                "_id": doc.get(id_key),
                "_source": doc
            } for doc in data
        ]
        success, errors = helpers.bulk(self.client, actions)
        if errors:
            self._log(f"Bulk insert encountered errors: {errors}", "error")
        else:
            self._log("Bulk insert completed successfully", "info")
        return success, errors




import httpx
import pytest
import vcr
from ppp_connectors.api_connectors.twilio import TwilioConnector

@pytest.mark.integration
def test_lookup_phone_vcr(vcr_cassette):
    with vcr_cassette.use_cassette("test_lookup_phone_vcr"):
        connector = TwilioConnector(load_env_vars=True)
        result = connector.lookup_phone("+14155552671")

        assert isinstance(result, httpx.Response)
        assert "phone_number" in result.json() or "caller_name" in result.json() or "line_type" in result.json()


import pytest
import httpx
from unittest.mock import patch, MagicMock
from ppp_connectors.dbms_connectors.splunk import SplunkConnector


@pytest.fixture
def connector():
    return SplunkConnector(
        host="localhost",
        port=8089,
        username="admin",
        password="admin123",
        verify=False
    )


@patch("httpx.post")
@patch("httpx.get")
def test_query_success(mock_get, mock_post, connector):
    # Mock POST /services/search/jobs
    post_response = MagicMock()
    post_response.raise_for_status = MagicMock()
    post_response.json.return_value = {"sid": "abc123"}
    mock_post.return_value = post_response

    # Mock GET /services/search/jobs/{sid}
    get_status_response = MagicMock()
    get_status_response.raise_for_status = MagicMock()
    get_status_response.json.return_value = {
        "entry": [{"content": {"isDone": True}}]
    }

    # Mock GET /services/search/jobs/{sid}/results
    get_results_response = MagicMock()
    get_results_response.raise_for_status = MagicMock()
    get_results_response.json.side_effect = [
        {"results": [{"foo": "bar"}, {"baz": "qux"}]},
        {"results": []}
    ]
    mock_get.side_effect = [get_status_response, get_results_response, get_results_response]

    results = list(connector.query("search index=_internal | head 2"))
    assert len(results) == 2
    assert results[0]["foo"] == "bar"
    assert results[1]["baz"] == "qux"


@patch("httpx.post")
def test_query_auth_error(mock_post, connector):
    mock_response = MagicMock()
    mock_response.raise_for_status.side_effect = httpx.HTTPError("401 Unauthorized")
    mock_post.return_value = mock_response

    with pytest.raises(httpx.HTTPError):
        list(connector.query("search index=_internal | head 1"))



import pytest
import httpx
from ppp_connectors.api_connectors.broker import AsyncBroker


def test_asyncbroker_rejects_mounts():
    """Ensure that AsyncBroker raises an error if 'mounts' is passed."""
    with pytest.raises(ValueError, match="mounts.*not supported"):
        AsyncBroker(base_url="https://example.com", mounts={"http://": object()})


def test_asyncbroker_explicit_proxy_used(mocker):
    """Ensure that an explicitly provided proxy is passed to the AsyncClient."""
    client_mock = mocker.patch("httpx.AsyncClient")
    AsyncBroker(base_url="https://example.com", proxy="http://explicit:1234")
    client_mock.assert_called_once()
    assert client_mock.call_args.kwargs["proxy"] == "http://explicit:1234"


import pytest
import httpx
from unittest.mock import patch, AsyncMock
from ppp_connectors.api_connectors.twilio import AsyncTwilioConnector

@pytest.mark.asyncio
async def test_async_init_with_env(monkeypatch):
    monkeypatch.setenv("TWILIO_API_SID", "sid")
    monkeypatch.setenv("TWILIO_API_SECRET", "secret")

    connector = AsyncTwilioConnector(load_env_vars=True)
    assert connector.api_sid == "sid"
    assert connector.api_secret == "secret"

@pytest.mark.asyncio
async def test_async_lookup_phone_raises_for_invalid_packages():
    connector = AsyncTwilioConnector(api_sid="a", api_secret="b")
    with pytest.raises(ValueError, match="Invalid data packages: badpkg"):
        await connector.lookup_phone("+14155552671", data_packages=["badpkg"])

@patch("ppp_connectors.api_connectors.twilio.AsyncTwilioConnector._make_request", new_callable=AsyncMock)
@pytest.mark.asyncio
async def test_async_lookup_phone_calls_make_request(mock_request):
    mock_resp = httpx.Response(200, content=b'{"valid": true}')
    mock_request.return_value = mock_resp

    connector = AsyncTwilioConnector(api_sid="a", api_secret="b")
    result = await connector.lookup_phone("+14155552671", data_packages=["caller_name"])

    assert result.status_code == 200
    assert mock_request.await_count == 1
    args, kwargs = mock_request.call_args
    assert kwargs["endpoint"].endswith("+14155552671")
    assert "caller_name" in kwargs["params"]["Fields"]


import httpx
from typing import Optional
from urllib.parse import quote
from ppp_connectors.api_connectors.broker import Broker, AsyncBroker, bubble_broker_init_signature, log_method_call

@bubble_broker_init_signature()
class IPQSConnector(Broker):
    """
    A connector for the IPQualityScore Malicious URL Scanner API.

    This class provides a typed interface to interact with IPQS's malicious URL
    scan endpoint. It handles API key management, header setup, and request routing
    through the shared Broker infrastructure.

    Attributes:
        api_key (str): The API key used to authenticate with IPQS.
    """
    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://ipqualityscore.com/api/json", **kwargs)

        self.api_key = api_key or self.env_config.get("IPQS_API_KEY")
        if not self.api_key:
            raise ValueError("API key is required for IPQSConnector")
        self.headers.update({"Content-Type": "application/json"})

    @log_method_call
    def malicious_url(self, query: str, **kwargs) -> httpx.Response:
        """
        Scan a URL using IPQualityScore's Malicious URL Scanner API.

        Args:
            query (str): The URL to scan (will be URL-encoded).
            **kwargs: Optional parameters like 'strictness' or 'fast' to influence scan behavior.

        Returns:
            httpx.Response: the httpx.Response object
        """
        encoded_query: str = quote(query, safe="")
        return self.post(f"/url/", json={"url": query, "key": self.api_key, **kwargs})


@bubble_broker_init_signature()
class AsyncIPQSConnector(AsyncBroker):
    """
    Async version of IPQSConnector using AsyncBroker infrastructure.
    """
    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://ipqualityscore.com/api/json", **kwargs)

        self.api_key = api_key or self.env_config.get("IPQS_API_KEY")
        if not self.api_key:
            raise ValueError("API key is required for AsyncIPQSConnector")
        self.headers.update({"Content-Type": "application/json"})

    @log_method_call
    async def malicious_url(self, query: str, **kwargs) -> httpx.Response:
        """
        Asynchronously scan a URL using IPQualityScore's Malicious URL Scanner API.

        Args:
            query (str): The URL to scan (will be URL-encoded).
            **kwargs: Optional parameters like 'strictness' or 'fast' to influence scan behavior.

        Returns:
            httpx.Response: the httpx.Response object
        """
        encoded_query: str = quote(query, safe="")
        return await self.post(f"/url/", json={"url": query, "key": self.api_key, **kwargs})



import pytest
import httpx
from unittest.mock import patch, AsyncMock
from ppp_connectors.api_connectors.ipqs import AsyncIPQSConnector


@pytest.mark.asyncio
async def test_async_init_with_api_key():
    connector = AsyncIPQSConnector(api_key="test_key")
    assert connector.api_key == "test_key"
    assert connector.headers["Content-Type"] == "application/json"


@pytest.mark.asyncio
async def test_async_init_with_env_key():
    with patch.dict("os.environ", {"IPQS_API_KEY": "env_key"}):
        connector = AsyncIPQSConnector(load_env_vars=True)
        assert connector.api_key == "env_key"
        assert connector.headers["Content-Type"] == "application/json"


@pytest.mark.asyncio
async def test_async_init_missing_key():
    with patch.dict("os.environ", {}, clear=True):
        with pytest.raises(ValueError, match="API key is required"):
            AsyncIPQSConnector()


@patch("ppp_connectors.api_connectors.ipqs.AsyncIPQSConnector.post", new_callable=AsyncMock)
@pytest.mark.asyncio
async def test_async_malicious_url(mock_post):
    import json

    request = httpx.Request("POST", "https://ipqualityscore.com/api/json/url/")
    payload = {"success": True, "domain": "example.com"}
    mock_response = httpx.Response(
        200,
        request=request,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_post.return_value = mock_response

    connector = AsyncIPQSConnector(api_key="test_key")
    response = await connector.malicious_url("example.com", strictness=1)

    assert isinstance(response, httpx.Response)
    assert response.status_code == 200
    assert response.json() == payload
    mock_post.assert_awaited_once_with(
        "/url/",
        json={"url": "example.com", "key": "test_key", "strictness": 1}
    )


import httpx
from typing import Optional
from ppp_connectors.api_connectors.broker import Broker, AsyncBroker, bubble_broker_init_signature, log_method_call

@bubble_broker_init_signature()
class FlashpointConnector(Broker):
    """
    FlashpointConnector provides access to various Flashpoint API search and retrieval endpoints
    using a consistent Broker-based interface.

    Attributes:
        api_key (str): Flashpoint API token used for bearer authentication.
    """
    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://api.flashpoint.io", **kwargs)
        self.api_key = api_key or self.env_config.get("FLASHPOINT_API_KEY")
        if not self.api_key:
            raise ValueError("FLASHPOINT_API_KEY is required")
        self.headers.update({
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        })

    @log_method_call
    def search_communities(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint communities data.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return self.post("/sources/v2/communities", json={"query": query, **kwargs})

    @log_method_call
    def search_fraud(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint fraud datasets.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return self.post("/sources/v2/fraud", json={"query": query, **kwargs})

    @log_method_call
    def search_marketplaces(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint marketplace datasets.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return self.post("/sources/v2/markets", json={"query": query, **kwargs})

    @log_method_call
    def search_media(self, query: str, **kwargs) -> httpx.Response:
        """
        Search OCR-processed media from Flashpoint.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return self.post("/sources/v2/media", json={"query": query, **kwargs})

    @log_method_call
    def get_media_object(self, query: str, **kwargs) -> httpx.Response:
        """
        Retrieve metadata for a specific media object.

        Args:
            query (str): The media_id of the object to retrieve.
            **kwargs: Additional request options.
        """
        return self.get(f"/sources/v2/media/{query}")

    @log_method_call
    def get_media_image(self, query: str, **kwargs) -> httpx.Response:
        """
        Download image asset by storage_uri.

        Args:
            query (str): The storage_uri (asset_id) of the image to download.
            **kwargs: Additional request options.
        """
        safe_headers = {"Authorization": f"Bearer {self.api_key}"}
        return self.get("/sources/v1/media/", headers=safe_headers, params={"asset_id": query})

    @log_method_call
    def search_checks(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint fraud check datasets.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return self.post("/sources/v2/fraud/checks", json={"query": query, **kwargs})


# Async version of FlashpointConnector
@bubble_broker_init_signature()
class AsyncFlashpointConnector(AsyncBroker):
    """
    AsyncFlashpointConnector provides async access to Flashpoint API endpoints.
    """
    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://api.flashpoint.io", **kwargs)
        self.api_key = api_key or self.env_config.get("FLASHPOINT_API_KEY")
        if not self.api_key:
            raise ValueError("FLASHPOINT_API_KEY is required")
        self.headers.update({
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        })

    @log_method_call
    async def search_communities(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint communities data.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return await self.post("/sources/v2/communities", json={"query": query, **kwargs})

    @log_method_call
    async def search_fraud(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint fraud datasets.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return await self.post("/sources/v2/fraud", json={"query": query, **kwargs})

    @log_method_call
    async def search_marketplaces(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint marketplace datasets.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return await self.post("/sources/v2/markets", json={"query": query, **kwargs})

    @log_method_call
    async def search_media(self, query: str, **kwargs) -> httpx.Response:
        """
        Search OCR-processed media from Flashpoint.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return await self.post("/sources/v2/media", json={"query": query, **kwargs})

    @log_method_call
    async def get_media_object(self, query: str) -> httpx.Response:
        """
        Retrieve metadata for a specific media object.

        Args:
            query (str): The media_id of the object to retrieve.
        """
        return await self.get(f"/sources/v2/media/{query}")

    @log_method_call
    async def get_media_image(self, query: str) -> httpx.Response:
        """
        Download image asset by storage_uri.

        Args:
            query (str): The storage_uri (asset_id) of the image to download.
        """
        safe_headers = {"Authorization": f"Bearer {self.api_key}"}
        return await self.get("/sources/v1/media/", headers=safe_headers, params={"asset_id": query})

    @log_method_call
    async def search_checks(self, query: str, **kwargs) -> httpx.Response:
        """
        Search Flashpoint fraud check datasets asynchronously.

        Args:
            query (str): The search string used in the API query.
            **kwargs: Additional query logic per the Flashpoint API documentation.
        """
        return await self.post("/sources/v2/fraud/checks", json={"query": query, **kwargs})


"""
Shim for backward compatibility.
Forwards to ppp_connectors.connectors.broker.
TODO: Deprecate this file in a future major release.
"""

import warnings
from ppp_connectors.connectors.broker import *

warnings.warn(
    "ppp_connectors.broker is deprecated and will be removed in a future release; use ppp_connectors.connectors.broker instead.",
    DeprecationWarning,
    stacklevel=2
)


import httpx
import pytest
from ppp_connectors.api_connectors.ipqs import IPQSConnector

@pytest.mark.integration
def test_ipqs_malicious_url_vcr(vcr_cassette):
    with vcr_cassette.use_cassette("test_ipqs_malicious_url_vcr"):
        connector = IPQSConnector(load_env_vars=True, enable_logging=True)
        result = connector.malicious_url("github.com")

        assert isinstance(result, httpx.Response)
        assert "domain" in result.json()
        assert result.json()["domain"] == "github.com"


import pytest
from unittest.mock import MagicMock, patch
from ppp_connectors.dbms_connectors.odbc import ODBCConnector


@patch("pyodbc.connect")
def test_odbcconnector_init(mock_connect):
    mock_logger = MagicMock()
    connector = ODBCConnector("DSN=testdb", logger=mock_logger)
    mock_connect.assert_called_once_with("DSN=testdb")
    assert connector.logger == mock_logger


@patch("pyodbc.connect")
def test_odbcconnector_query_returns_rows(mock_connect):
    """Test that query returns rows as dictionaries."""
    mock_cursor = MagicMock()
    mock_cursor.description = [("id",), ("name",)]
    mock_cursor.fetchall.return_value = [(1, "Alice"), (2, "Bob")]
    mock_connect.return_value.cursor.return_value = mock_cursor
    connector = ODBCConnector("DSN=testdb")

    results = list(connector.query("SELECT * FROM users"))

    assert results == [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]


@patch("pyodbc.connect")
def test_odbcconnector_bulk_insert(mock_connect):
    mock_cursor = MagicMock()
    mock_connect.return_value.cursor.return_value = mock_cursor
    connector = ODBCConnector("DSN=testdb")

    data = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    connector.bulk_insert("users", data)

    assert mock_cursor.executemany.called
    assert mock_connect.return_value.commit.called


@patch("pyodbc.connect")
def test_odbcconnector_bulk_insert_empty_data(mock_connect):
    mock_cursor = MagicMock()
    mock_connect.return_value.cursor.return_value = mock_cursor
    connector = ODBCConnector("DSN=testdb")

    connector.bulk_insert("users", [])

    mock_cursor.executemany.assert_not_called()


@patch("pyodbc.connect")
def test_odbcconnector_context_manager_closes_connection(mock_connect):
    """Test that the context manager closes the connection."""
    mock_conn = MagicMock()
    mock_connect.return_value = mock_conn

    with ODBCConnector("DSN=testdb") as connector:
        assert isinstance(connector, ODBCConnector)

    mock_conn.close.assert_called_once()


import httpx
from typing import Dict, Any, Optional
from ppp_connectors.api_connectors.broker import Broker, AsyncBroker, bubble_broker_init_signature, log_method_call

@bubble_broker_init_signature()
class GenericConnector(Broker):
    """
    A flexible, minimal connector that allows sending arbitrary HTTP requests
    using the Broker infrastructure.
    """

    @log_method_call
    def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Any] = None,
        json: Optional[Dict[str, Any]] = None,
        auth: Optional[Any] = None,
        timeout: Optional[int] = None,
        **kwargs
    ) -> httpx.Response:
        """
        Make an arbitrary HTTP request using Broker's request logic.

        Args:
            method (str): HTTP method (e.g., GET, POST).
            url (str): Fully qualified URL to send the request to.
            headers (Optional[Dict[str, str]]): Request headers.
            params (Optional[Dict[str, Any]]): Query string parameters.
            data (Optional[Any]): Form-encoded or raw data.
            json (Optional[Dict[str, Any]]): JSON payload.
            auth (Optional[Any]): Authentication (e.g., tuple for basic auth).
            timeout (Optional[int]): Request timeout in seconds.

        Returns:
            httpx.Response: The response object.
        """

        # Merge headers with base class
        merged_headers = self.headers.copy()
        if headers:
            merged_headers.update(headers)

        return self._make_request(
            method=method,
            endpoint=url,
            headers=merged_headers,
            params=params,
            json=json,
            auth=auth,
            retry_kwargs=kwargs.get("retry_kwargs"),
        )


@bubble_broker_init_signature()
class AsyncGenericConnector(AsyncBroker):
    """
    Async version of GenericConnector using AsyncBroker infrastructure.
    """

    @log_method_call
    async def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Any] = None,
        json: Optional[Dict[str, Any]] = None,
        auth: Optional[Any] = None,
        timeout: Optional[int] = None,
        **kwargs
    ) -> httpx.Response:
        """
        Make an arbitrary HTTP request using AsyncBroker's request logic.

        Args:
            method (str): HTTP method (e.g., GET, POST).
            url (str): Fully qualified URL to send the request to.
            headers (Optional[Dict[str, str]]): Request headers.
            params (Optional[Dict[str, Any]]): Query string parameters.
            data (Optional[Any]): Form-encoded or raw data.
            json (Optional[Dict[str, Any]]): JSON payload.
            auth (Optional[Any]): Authentication (e.g., tuple for basic auth).
            timeout (Optional[int]): Request timeout in seconds.

        Returns:
            httpx.Response: The response object.
        """
        merged_headers = self.headers.copy()
        if headers:
            merged_headers.update(headers)

        return await self._make_request(
            method=method,
            endpoint=url,
            headers=merged_headers,
            params=params,
            json=json,
            auth=auth,
            retry_kwargs=kwargs.get("retry_kwargs"),
        )



import json
import httpx
import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.spycloud import SpycloudConnector


def test_init_with_api_key():
    connector = SpycloudConnector(sip_key="sip", ato_key="ato", inv_key="inv")
    assert connector.sip_key == "sip"
    assert connector.ato_key == "ato"
    assert connector.inv_key == "inv"


@patch.dict("os.environ", {"SPYCLOUD_API_ATO_KEY": "env_key"}, clear=True)
def test_init_with_env_key():
    connector = SpycloudConnector(load_env_vars=True)
    assert connector.ato_key == "env_key"


@patch("ppp_connectors.api_connectors.broker.combine_env_configs", return_value={})
def test_init_missing_key(mock_env):
    connector = SpycloudConnector(load_env_vars=True)
    with pytest.raises(ValueError, match="SPYCLOUD_API_ATO_KEY is required for this request"):
        connector.ato_search(search_type="email", query="test@example.com")


@patch("ppp_connectors.api_connectors.spycloud.SpycloudConnector._make_request")
def test_ato_search(mock_request):
    # Build a real httpx.Response to match new return type (raw Response)
    request = httpx.Request("GET", "https://api.spycloud.io/sp-v2/breach/data/emails/test@example.com")
    payload = {"results": []}
    mock_response = httpx.Response(
        200,
        request=request,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_request.return_value = mock_response

    connector = SpycloudConnector(ato_key="test_key")
    result = connector.ato_search(search_type="ip", query="test@example.com")

    mock_request.assert_called_once()
    assert isinstance(result, httpx.Response)
    assert result.json() == payload


import pytest
from httpx import Response
from httpx import Request
from ppp_connectors.api_connectors.urlscan import AsyncURLScanConnector

@pytest.mark.asyncio
async def test_async_init_with_api_key():
    connector = AsyncURLScanConnector(api_key="testkey")
    assert connector.api_key == "testkey"
    assert connector.headers["API-Key"] == "testkey"

@pytest.mark.asyncio
async def test_async_init_with_env(monkeypatch):
    monkeypatch.setenv("URLSCAN_API_KEY", "envkey")
    connector = AsyncURLScanConnector(load_env_vars=True)
    assert connector.api_key == "envkey"
    assert connector.headers["API-Key"] == "envkey"

@pytest.mark.asyncio
async def test_async_init_missing_key_raises(monkeypatch):
    monkeypatch.delenv("URLSCAN_API_KEY", raising=False)
    with pytest.raises(ValueError, match="API key is required for AsyncURLScanConnector"):
        AsyncURLScanConnector()

@pytest.mark.asyncio
async def test_async_search_makes_expected_call(monkeypatch):
    connector = AsyncURLScanConnector(api_key="testkey")

    async def mock_get(path, params=None):
        assert path == "/api/v1/search/"
        assert params["q"] == "example.com"
        return Response(200, request=Request("GET", path))

    connector.get = mock_get
    resp = await connector.search("example.com")
    assert resp.status_code == 200

@pytest.mark.asyncio
async def test_async_scan_makes_expected_call(monkeypatch):
    connector = AsyncURLScanConnector(api_key="testkey")

    async def mock_post(path, json=None):
        assert path == "/api/v1/scan"
        assert json["url"] == "http://example.com"
        return Response(200, request=Request("POST", path))

    connector.post = mock_post
    resp = await connector.scan("http://example.com")
    assert resp.status_code == 200


import httpx
from typing import Dict, Any, Optional
from ppp_connectors.api_connectors.broker import AsyncBroker, Broker, bubble_broker_init_signature, log_method_call

@bubble_broker_init_signature()
class URLScanConnector(Broker):
    """
    A connector for interacting with the urlscan.io API.

    Provides structured methods for submitting scans, querying historical data,
    and retrieving detailed scan results and metadata.

    Attributes:
        api_key (str): The API key used to authenticate with urlscan.io.
    """

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://urlscan.io", **kwargs)

        self.api_key = api_key or self.env_config.get("URLSCAN_API_KEY")
        if not self.api_key:
            raise ValueError("API key is required for URLScanConnector")
        self.headers.update({
            "accept": "application/json",
            "API-Key": self.api_key
        })

    @log_method_call
    def search(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Search for archived scans matching a given query.

        Args:
            query (str): The search term or filter string.
            **kwargs: Additional query parameters for filtering results.

        Returns:
            httpx.Response: the httpx.Response object
        """
        params = {"q": query, **kwargs}
        return self.get("/api/v1/search/", params=params)

    @log_method_call
    def scan(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Submit a URL to be scanned by urlscan.io.

        Args:
            query (str): The URL to scan.
            **kwargs: Additional scan options like tags, visibility, or referer.

        Returns:
            httpx.Response: the httpx.Response object
        """
        payload = {"url": query, **kwargs}
        return self.post("/api/v1/scan", json=payload)

    @log_method_call
    def results(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Retrieve detailed scan results by UUID.

        Args:
            query (str): The UUID of the scan.

        Returns:
            httpx.Response: the httpx.Response object
        """
        return self.get(f"/api/v1/result/{query}", params=kwargs)

    @log_method_call
    def get_dom(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Retrieve the DOM snapshot for a given scan UUID.

        Args:
            query (str): The UUID of the scan.

        Returns:
            httpx.Response: the httpx.Response object
        """
        return self.get(f"/dom/{query}", params=kwargs)

    @log_method_call
    def structure_search(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Search for scans structurally similar to a given UUID.

        Args:
            query (str): The UUID of the original scan.

        Returns:
            httpx.Response: the httpx.Response object
        """
        return self.get(f"/api/v1/pro/result/{query}/similar", params=kwargs)

class AsyncURLScanConnector(AsyncBroker):
    """
    An async connector for interacting with the urlscan.io API.
    """

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://urlscan.io", **kwargs)

        self.api_key = api_key or self.env_config.get("URLSCAN_API_KEY")
        if not self.api_key:
            raise ValueError("API key is required for AsyncURLScanConnector")
        self.headers.update({
            "accept": "application/json",
            "API-Key": self.api_key
        })

    @log_method_call
    async def search(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Async search for archived scans matching a given query.
        """
        params = {"q": query, **kwargs}
        return await self.get("/api/v1/search/", params=params)

    @log_method_call
    async def scan(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Async submit a URL to be scanned by urlscan.io.
        """
        payload = {"url": query, **kwargs}
        return await self.post("/api/v1/scan", json=payload)

    @log_method_call
    async def results(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Async retrieve detailed scan results by UUID.
        """
        return await self.get(f"/api/v1/result/{query}", params=kwargs)

    @log_method_call
    async def get_dom(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Async retrieve the DOM snapshot for a given scan UUID.
        """
        return await self.get(f"/dom/{query}", params=kwargs)

    @log_method_call
    async def structure_search(self, query: str, **kwargs: Dict[str, Any]) -> httpx.Response:
        """
        Async search for scans structurally similar to a given UUID.
        """
        return await self.get(f"/api/v1/pro/result/{query}/similar", params=kwargs)



import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.generic import GenericConnector


def test_init_sets_base_url():
    connector = GenericConnector(base_url="https://example.com")
    assert connector.base_url == "https://example.com"


@patch("ppp_connectors.api_connectors.generic.GenericConnector._make_request")
def test_get_request(mock_make_request):
    mock_response = MagicMock()
    mock_response.json.return_value = {"result": "ok"}
    mock_make_request.return_value = mock_response

    connector = GenericConnector(base_url="https://example.com")
    response = connector.get("/test")

    assert response.json() == {"result": "ok"}
    mock_make_request.assert_called_once_with("GET", "/test", params=None)


@patch("ppp_connectors.api_connectors.generic.GenericConnector._make_request")
def test_post_request(mock_make_request):
    mock_response = MagicMock()
    mock_response.json.return_value = {"posted": True}
    mock_make_request.return_value = mock_response

    connector = GenericConnector(base_url="https://example.com")
    response = connector.post("/submit", json={"key": "value"})

    assert response.json() == {"posted": True}
    mock_make_request.assert_called_once_with("POST", "/submit", json={"key": "value"})



import httpx
import pytest
from ppp_connectors.api_connectors.flashpoint import FlashpointConnector

@pytest.mark.integration
def test_flashpoint_search_fraud_vcr(vcr_cassette):
    with vcr_cassette.use_cassette("test_flashpoint_search_fraud_vcr"):
        connector = FlashpointConnector(load_env_vars=True)
        result = connector.search_fraud("dark web")
        assert isinstance(result, httpx.Response)
        assert "items" in result.json()


import pytest
from pymongo import UpdateOne
from pymongo.errors import ServerSelectionTimeoutError
from ppp_connectors.dbms_connectors.mongo import MongoConnector
from unittest.mock import patch, MagicMock


def test_mongo_query(monkeypatch):
    # Prevent real connection by mocking MongoClient and returning a successful ping
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )
    mock_cursor = [
        {"_id": 1, "name": "Alice"},
        {"_id": 2, "name": "Bob"}
    ]
    mock_collection = MagicMock()
    mock_collection.find.return_value.batch_size.return_value = mock_cursor
    mock_db = {"test_collection": mock_collection}
    mock_client = {"test_db": mock_db}

    connector = MongoConnector(
        uri="mongodb://localhost:27017",
        username="fake",
        password="fake",
    )
    connector.client = mock_client

    results = list(connector.query("test_db", "test_collection", {}))
    assert results == mock_cursor


def test_insert_many(monkeypatch):
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )
    mock_insert_many = MagicMock()
    mock_collection = {"insert_many": mock_insert_many}
    mock_db = {"test_collection": MagicMock(return_value=mock_insert_many)}
    mock_client = {"test_db": mock_db}

    connector = MongoConnector(
        uri="mongodb://localhost:27017",
        username="fake",
        password="fake",
    )
    connector.client = MagicMock()
    connector.client.__getitem__.return_value.__getitem__.return_value.insert_many = mock_insert_many

    test_data = [{"_id": 1}, {"_id": 2}]
    connector.insert_many("test_db", "test_collection", test_data)

    mock_insert_many.assert_called_once_with(test_data, ordered=False)


def test_distinct(monkeypatch):
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )
    mock_collection = MagicMock()
    mock_collection.distinct.return_value = ["A", "B"]
    connector = MongoConnector(
        uri="mongodb://localhost:27017",
        username="fake",
        password="fake",
        ssl=False
    )
    connector.client = MagicMock()
    connector.client.__getitem__.return_value.__getitem__.return_value = mock_collection

    values = connector.distinct("test_db", "test_collection", key="field", filter={"x": 1}, maxTimeMS=5000)
    assert values == ["A", "B"]
    mock_collection.distinct.assert_called_once_with("field", {"x": 1}, maxTimeMS=5000)


def test_mongo_connection_failure():
    with pytest.raises(ServerSelectionTimeoutError):
        MongoConnector(
            uri="mongodb://localhost:27018",  # invalid port
            username="fake",
            password="fake",
            timeout=1,
        )



@patch("ppp_connectors.dbms_connectors.mongo.MongoClient")
def test_mongo_init_with_auth_and_ssl(mock_mongo_client):
    # Mock ping success on the returned client instance
    instance = MagicMock()
    instance.admin.command.return_value = {"ok": 1}
    mock_mongo_client.return_value = instance

    MongoConnector(
        uri="mongodb://example.com:27017",
        username="user",
        password="pass",
        auth_source="authdb",
        auth_mechanism="SCRAM-SHA-1",
        ssl=True
    )
    mock_mongo_client.assert_called_once_with(
        "mongodb://example.com:27017",
        username="user",
        password="pass",
        authSource="authdb",
        authMechanism="SCRAM-SHA-1",
        ssl=True,
        serverSelectionTimeoutMS=10000
    )


@patch("ppp_connectors.dbms_connectors.mongo.MongoClient")
def test_mongo_init_defaults(mock_mongo_client):
    instance = MagicMock()
    instance.admin.command.return_value = {"ok": 1}
    mock_mongo_client.return_value = instance
    MongoConnector(uri="mongodb://localhost:27017")
    mock_mongo_client.assert_called_once()



def test_upsert_many_with_unique_key(monkeypatch):
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )
    mock_bulk_write = MagicMock()
    mock_collection = MagicMock()
    mock_collection.bulk_write = mock_bulk_write

    connector = MongoConnector(
        uri="mongodb://localhost:27017",
        username="fake",
        password="fake",
    )
    connector.client = MagicMock()
    connector.client.__getitem__.return_value.__getitem__.return_value = mock_collection

    data = [{"_id": 1, "value": "A"}, {"_id": 2, "value": "B"}]
    connector.upsert_many("test_db", "test_collection", data, unique_key="_id")

    ops = [
        UpdateOne({"_id": 1}, {"$set": {"_id": 1, "value": "A"}}, upsert=True),
        UpdateOne({"_id": 2}, {"$set": {"_id": 2, "value": "B"}}, upsert=True)
    ]
    mock_bulk_write.assert_called_once_with(ops, ordered=False)


def test_upsert_many_raises_without_unique_key(monkeypatch):
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )
    connector = MongoConnector(
        uri="mongodb://localhost:27017",
        username="fake",
        password="fake",
    )
    mock_collection = MagicMock()
    mock_collection.bulk_write = MagicMock()
    connector.client = MagicMock()
    connector.client.__getitem__.return_value.__getitem__.return_value = mock_collection

    data = [{"_id": i} for i in range(10)]
    with pytest.raises(ValueError, match="unique_key must be provided for upsert_many"):
        connector.upsert_many("test_db", "test_collection", data, unique_key=None)


def test_context_manager_closes_client(monkeypatch):
    fake_client = MagicMock()
    fake_client.admin.command.return_value = {"ok": 1}
    monkeypatch.setattr(
        "ppp_connectors.dbms_connectors.mongo.MongoClient",
        MagicMock(return_value=fake_client),
    )

    with MongoConnector(uri="mongodb://localhost:27017") as conn:
        assert isinstance(conn, MongoConnector)

    # After exiting context, close should be called
    fake_client.close.assert_called_once()



from pymongo import MongoClient, UpdateOne
from pymongo.errors import (
    OperationFailure,
    ServerSelectionTimeoutError,
    AutoReconnect,
    ConnectionFailure,
)
from tenacity import Retrying, stop_after_attempt, wait_fixed, retry_if_exception_type
from typing import List, Dict, Any, Optional, Generator, Type
from types import TracebackType
from ppp_connectors.helpers import setup_logger


class MongoConnector:
    """
    A connector class for interacting with MongoDB.

    Provides methods for finding documents with paging and performing batched
    insert and upsert operations, as well as convenience helpers for
    `distinct`, `delete`, and `delete_many`.

    Supports explicit lifecycle management via `close()` and can be used as a
    context manager (`with MongoConnector(...) as conn:`). On initialization,
    the connector pings the server to validate connectivity/authentication with
    a simple retry policy.
    Logs actions if a logger is provided.
    """
    def __init__(
        self,
        uri: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        auth_source: str = "admin",
        timeout: int = 10,
        auth_mechanism: Optional[str] = "DEFAULT",
        ssl: Optional[bool] = True,
        logger: Optional[Any] = None,
        auth_retry_attempts: int = 3,
        auth_retry_wait: float = 1.0,
    ):
        """
        Initialize the MongoDB client.

        Args:
            uri (str): The MongoDB connection URI.
            username (Optional[str]): Username for authentication. Defaults to None.
            password (Optional[str]): Password for authentication. Defaults to None.
            auth_source (str): The authentication database. Defaults to "admin".
            timeout (int): Server selection timeout in seconds. Defaults to 10.
            auth_mechanism (Optional[str]): Authentication mechanism for MongoDB (e.g., "SCRAM-SHA-1").
            ssl (Optional[bool]): Whether to use SSL for the connection.
            logger (Optional[Any]): Logger instance for logging actions. Defaults to None.
            auth_retry_attempts (int): Number of attempts for initial auth ping. Defaults to 3.
            auth_retry_wait (float): Seconds to wait between auth attempts. Defaults to 1.0.
        """
        # Initialize MongoClient with authSource, authMechanism, and ssl options
        self.client = MongoClient(
            uri,
            username=username,
            password=password,
            authSource=auth_source,
            authMechanism=auth_mechanism,
            ssl=ssl,
            serverSelectionTimeoutMS=timeout * 1000
        )
        self.logger = logger or setup_logger(__name__)
        self.auth_retry_attempts = auth_retry_attempts
        self.auth_retry_wait = auth_retry_wait
        self._log(
            f"Initialized MongoClient with authSource={auth_source}, "
            f"authMechanism={auth_mechanism}, ssl={ssl}"
        )
        # Force an initial ping to trigger auth/handshake; retry to handle
        # clusters that intermittently fail the first attempt.
        self._ping_with_retry()

    def close(self) -> None:
        """Close the underlying MongoClient connection."""
        self._log("Closing MongoClient connection", level="debug")
        try:
            self.client.close()
        except Exception as exc:
            # Closing should be best-effort; log and continue
            self._log(f"Error during MongoClient.close(): {exc}", level="warning")

    # Context manager support
    def __enter__(self) -> "MongoConnector":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def _log(self, msg: str, level: str = "info"):
        """
        Internal helper method for logging.

        Args:
            msg (str): The message to log.
            level (str): Logging level as string (e.g., "info", "debug"). Defaults to "info".
        """
        if self.logger:
            log_method = getattr(self.logger, level, self.logger.info)
            log_method(msg)

    def _ping_with_retry(self) -> None:
        """Ping the server to validate connection/auth, with retry."""
        for attempt in Retrying(
            stop=stop_after_attempt(self.auth_retry_attempts),
            wait=wait_fixed(self.auth_retry_wait),
            reraise=True,
            retry=retry_if_exception_type(
                (OperationFailure, ServerSelectionTimeoutError, AutoReconnect, ConnectionFailure)
            ),
        ):
            with attempt:
                self._log("Pinging MongoDB to verify connection/auth...", level="debug")
                # 'ping' triggers handshake and, when needed, authentication
                self.client.admin.command("ping")

    def find(
        self,
        db_name: str,
        collection: str,
        filter: Dict,
        projection: Optional[Dict] = None,
        batch_size: int = 1000
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Find documents in a MongoDB collection with optional projection and paging.

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            filter (Dict): MongoDB filter document.
            projection (Optional[Dict]): Fields to include or exclude. Defaults to None.
            batch_size (int): Number of documents per batch. Defaults to 1000.

        Yields:
            Dict[str, Any]: Each document as a dictionary.

        Logs:
            Logs the find operation with filter details.
        """
        self._log(f"Executing Mongo find on {db_name}.{collection} with filter: {filter}")
        col = self.client[db_name][collection]
        cursor = col.find(filter, projection).batch_size(batch_size)
        for doc in cursor:
            yield doc

    def query(
        self,
        db_name: str,
        collection: str,
        query: Dict,
        projection: Optional[Dict] = None,
        batch_size: int = 1000
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Deprecated: use `find` instead.

        Backwards-compatible wrapper that forwards to `find`.
        """
        self._log(
            "MongoConnector.query is deprecated; use MongoConnector.find instead",
            level="warning",
        )
        return self.find(
            db_name=db_name,
            collection=collection,
            filter=query,
            projection=projection,
            batch_size=batch_size,
        )

    def insert_many(
        self,
        db_name: str,
        collection: str,
        data: List[Dict],
        ordered: bool = False,
        batch_size: int = 1000,
    ):
        """
        Insert multiple documents into a collection (batched).

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            data (List[Dict]): Documents to insert.
            ordered (bool): Whether operations should be ordered. Defaults to False.
            batch_size (int): Number of documents per batch. Defaults to 1000.

        Returns:
            List: List of InsertManyResult objects for each batch.

        Note:
            PyMongo batches writes internally; manual batching here is useful
            for memory control, error isolation, and progress logging.
        """
        self._log(
            f"insert_many: inserting {len(data)} docs into {db_name}.{collection} with batch_size={batch_size}"
        )
        col = self.client[db_name][collection]
        results = []
        for i in range(0, len(data), batch_size):
            batch = data[i:i + batch_size]
            result = col.insert_many(batch, ordered=ordered)
            results.append(result)
        return results

    def upsert_many(
        self,
        db_name: str,
        collection: str,
        data: List[Dict],
        unique_key: Optional[str],
        ordered: bool = False,
        batch_size: int = 1000,
    ):
        """
        Upsert multiple documents into a collection using a unique key (batched).

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            data (List[Dict]): Documents to upsert.
            unique_key (Optional[str]): Field to use for upsert filtering. Required.
            ordered (bool): Whether operations should be ordered. Defaults to False.
            batch_size (int): Number of documents per batch. Defaults to 1000.

        Returns:
            List: List of BulkWriteResult objects for each batch.

        Details:
            Uses `bulk_write` with `UpdateOne({unique_key: value}, {"$set": doc}, upsert=True)`
            to merge fields from each document. The `unique_key` must exist in each document.
        """
        if not unique_key:
            raise ValueError("unique_key must be provided for upsert_many")
        self._log(
            f"upsert_many: upserting {len(data)} docs into {db_name}.{collection} with batch_size={batch_size}, unique_key={unique_key}"
        )
        col = self.client[db_name][collection]
        results = []
        for i in range(0, len(data), batch_size):
            batch = data[i:i + batch_size]
            operations = [
                UpdateOne({unique_key: doc[unique_key]}, {"$set": doc}, upsert=True)
                for doc in batch if unique_key in doc
            ]
            if operations:
                result = col.bulk_write(operations, ordered=ordered)
                results.append(result)
        return results

    def distinct(
        self,
        db_name: str,
        collection: str,
        key: str,
        filter: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Any]:
        """
        Return a list of distinct values for `key` across documents.

        This is a thin wrapper around PyMongo's `Collection.distinct` and accepts
        any additional keyword arguments supported by PyMongo (e.g., collation,
        maxTimeMS, comment).

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            key (str): Field name for which to return distinct values.
            filter (Optional[Dict[str, Any]]): Optional query filter to limit the scope.
            **kwargs: Additional options forwarded to `Collection.distinct`.

        Returns:
            List[Any]: Distinct values for the specified key.
        """
        self._log(
            f"Executing Mongo distinct on {db_name}.{collection} for key='{key}' with filter: {filter or {}}"
        )
        col = self.client[db_name][collection]
        return col.distinct(key, filter, **kwargs)

    def delete(
        self,
        db_name: str,
        collection: str,
        filter: Dict[str, Any],
        **kwargs: Any,
    ):
        """
        Delete a single document matching the filter (wrapper over delete_one).

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            filter (Dict[str, Any]): Query filter selecting the document to delete.
            **kwargs: Additional options forwarded to `Collection.delete_one` (e.g., collation, comment).

        Returns:
            DeleteResult: The result of the delete operation.
        """
        self._log(
            f"Deleting one from {db_name}.{collection} with filter: {filter}",
            level="info",
        )
        col = self.client[db_name][collection]
        return col.delete_one(filter, **kwargs)

    def delete_many(
        self,
        db_name: str,
        collection: str,
        filter: Dict[str, Any],
        **kwargs: Any,
    ):
        """
        Delete all documents matching the filter (wrapper over delete_many).

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            filter (Dict[str, Any]): Query filter selecting documents to delete.
            **kwargs: Additional options forwarded to `Collection.delete_many` (e.g., collation, comment).

        Returns:
            DeleteResult: The result of the delete operation.
        """
        self._log(
            f"Deleting many from {db_name}.{collection} with filter: {filter}",
            level="info",
        )
        col = self.client[db_name][collection]
        return col.delete_many(filter, **kwargs)



import pytest
import vcr
from pathlib import Path

AUTH_PARAM_REDACT = [
    # List of substrings that, if present in a key, will cause redaction (case-insensitive)
    "key", "api_key", "access_token", "auth", "authorization", "user", "pass", "api", "x-api", "x_api"
]


# Redact sensitive values from request body parameters
import json
def redact_sensitive(request):
    """Redact sensitive values from request body parameters."""
    if request.body:
        try:
            body = json.loads(request.body)
            for k in list(body.keys()):
                if any(redact_key.lower() in k.lower() for redact_key in AUTH_PARAM_REDACT):
                    body[k] = "REDACTED"
            request.body = json.dumps(body)
        except Exception:
            pass  # skip non-JSON or malformed bodies

    if request.headers:
        for k in list(request.headers.keys()):
            if any(redact_key.lower() in k.lower() for redact_key in AUTH_PARAM_REDACT):
                request.headers[k] = "REDACTED"

    return request


@pytest.fixture
def vcr_cassette(request):
    """Provides a VCR instance that stores cassettes in a test-local 'cassettes' folder."""
    test_dir = Path(request.module.__file__).parent
    cassette_dir = test_dir / "cassettes"

    config = {
        "cassette_library_dir": str(cassette_dir),
        "path_transformer": vcr.VCR.ensure_suffix(".yaml"),
        "record_mode": "once",
        "filter_headers": [("Authorization", "DUMMY")],
        "before_record": redact_sensitive
    }

    return vcr.VCR(**config)


import unittest
from unittest.mock import patch, MagicMock
from ppp_connectors.dbms_connectors.elasticsearch import ElasticsearchConnector

class TestElasticsearchConnector(unittest.TestCase):

    def setUp(self):
        self.mock_logger = MagicMock()
        self.connector = ElasticsearchConnector(
            hosts=["http://localhost:9200"],
            username="user",
            password="pass",
            logger=self.mock_logger
        )

    @patch("ppp_connectors.dbms_connectors.elasticsearch.Elasticsearch")
    def test_initialization(self, mock_es):
        ElasticsearchConnector(
            hosts=["http://localhost:9200"],
            username="user",
            password="pass",
            logger=self.mock_logger
        )
        mock_es.assert_called_with(["http://localhost:9200"], basic_auth=("user", "pass"))

    @patch("ppp_connectors.dbms_connectors.elasticsearch.Elasticsearch")
    def test_query_scroll(self, mock_es):
        mock_client = MagicMock()
        mock_client.search.return_value = {
            "_scroll_id": "abc123",
            "hits": {"hits": [{"_id": 1}, {"_id": 2}]}
        }
        mock_client.scroll.side_effect = [
            {"_scroll_id": "abc123", "hits": {"hits": [{"_id": 3}]}},
            {"_scroll_id": "abc123", "hits": {"hits": []}},
        ]
        mock_es.return_value = mock_client
        connector = ElasticsearchConnector(
            hosts=["http://localhost:9200"],
            username="user",
            password="pass",
            logger=self.mock_logger
        )
        results = list(connector.query(index="test", query={"match_all": {}}))
        self.assertEqual(len(results), 3)
        self.assertEqual(results[0]["_id"], 1)

    @patch("ppp_connectors.dbms_connectors.elasticsearch.helpers.bulk")
    def test_bulk_insert(self, mock_bulk):
        mock_bulk.return_value = (3, [])
        data = [{"_id": "1", "name": "Alice"}, {"_id": "2", "name": "Bob"}, {"_id": "3", "name": "Charlie"}]
        success, errors = self.connector.bulk_insert(index="test-index", data=data)
        self.assertEqual(success, 3)
        self.assertEqual(errors, [])
        self.mock_logger.info.assert_called_with("Bulk insert completed successfully")

    @patch("ppp_connectors.dbms_connectors.elasticsearch.helpers.bulk")
    def test_bulk_insert_with_errors(self, mock_bulk):
        mock_bulk.return_value = (2, [{"error": "failed to insert"}])
        data = [{"_id": "1", "name": "Alice"}, {"_id": "2", "name": "Bob"}]
        success, errors = self.connector.bulk_insert(index="test-index", data=data)
        self.assertEqual(success, 2)
        self.assertTrue(errors)
        self.mock_logger.error.assert_called()

if __name__ == "__main__":
    unittest.main()


import httpx
import pytest
from ppp_connectors.api_connectors.spycloud import SpycloudConnector

@pytest.mark.integration
def test_spycloud_ato_search_vcr(vcr_cassette):
    with vcr_cassette.use_cassette("test_spycloud_ato_search_vcr"):
        connector = SpycloudConnector(load_env_vars=True, enable_logging=True)
        result = connector.ato_search(search_type="ip", query="8.8.8.8")

        assert isinstance(result, httpx.Response)
        assert "results" in result.json()


import httpx
from ppp_connectors.api_connectors.broker import Broker

def test_integration_get_with_mock_transport():
    def handler(request):
        assert request.url.path == "/hello"
        return httpx.Response(200, json={"msg": "ok"})

    transport = httpx.MockTransport(handler)
    broker = Broker(base_url="https://testserver", timeout=5)
    broker.session = httpx.Client(transport=transport)
    response = broker.get("/hello")
    assert response.status_code == 200
    assert response.json() == {"msg": "ok"}



import httpx
import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.twilio import TwilioConnector


def test_init_with_all_keys():
    connector = TwilioConnector(api_sid="sid", api_secret="secret")
    assert connector.api_sid == "sid"
    assert connector.auth is not None


@patch.dict("os.environ", {"TWILIO_API_SID": "sid", "TWILIO_API_SECRET": "secret"}, clear=True)
def test_init_with_env_keys():
    connector = TwilioConnector(load_env_vars=True)
    assert connector.api_sid == "sid"
    assert connector.auth is not None


@patch("ppp_connectors.api_connectors.broker.combine_env_configs", return_value={})
def test_init_missing_auth_keys(mock_env):
    with pytest.raises(ValueError, match="TWILIO_API_SID and TWILIO_API_SECRET are required"):
        TwilioConnector(load_env_vars=True)


@patch("ppp_connectors.api_connectors.twilio.TwilioConnector._make_request")
def test_lookup_phone(mock_request):
    import json
    # Return a real httpx.Response to reflect new return type
    req = httpx.Request("GET", "https://lookups.twilio.com/v2/PhoneNumbers/15555555555")
    payload = {"carrier": "example"}
    resp = httpx.Response(
        200,
        request=req,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_request.return_value = resp

    connector = TwilioConnector(api_sid="sid", api_secret="secret")
    result = connector.lookup_phone("15555555555")

    assert isinstance(result, httpx.Response)
    assert result.json() == payload
    mock_request.assert_called_once()


import pytest
import vcr
from ppp_connectors.api_connectors.generic import GenericConnector

@pytest.mark.integration
def test_generic_get_github_api(vcr_cassette):
    with vcr_cassette.use_cassette("test_generic_get_github_api"):
        connector = GenericConnector(base_url="https://api.github.com")
        response = connector.get("/")

        data = response.json()
        assert "current_user_url" in data



from __future__ import annotations
from typing import Any, Dict, List, Optional, AsyncIterator, Type
from types import TracebackType
import inspect
from pymongo import UpdateOne
from pymongo.errors import (
    OperationFailure,
    ServerSelectionTimeoutError,
    AutoReconnect,
    ConnectionFailure,
)
from tenacity import AsyncRetrying, stop_after_attempt, wait_fixed, retry_if_exception_type
from ppp_connectors.helpers import setup_logger


class AsyncMongoConnector:
    """
    An asyncio connector for interacting with MongoDB using PyMongo's async client.

    Mirrors the synchronous MongoConnector API with async methods and context
    management (`async with AsyncMongoConnector(...) as conn:`). On entry, the
    connector pings the server with a simple retry policy to validate the
    connection and trigger authentication.

    Args:
        uri (str): The MongoDB connection URI.
        username (Optional[str]): Username for authentication. Defaults to None.
        password (Optional[str]): Password for authentication. Defaults to None.
        auth_source (str): The authentication database. Defaults to "admin".
        timeout (int): Server selection timeout in seconds. Defaults to 10.
        auth_mechanism (Optional[str]): Authentication mechanism for MongoDB (e.g., "SCRAM-SHA-1").
        ssl (Optional[bool]): Whether to use SSL for the connection.
        logger (Optional[Any]): Logger instance for logging actions. Defaults to None.
        auth_retry_attempts (int): Number of attempts for initial auth ping. Defaults to 3.
        auth_retry_wait (float): Seconds to wait between auth attempts. Defaults to 1.0.
    """

    def __init__(
        self,
        uri: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        auth_source: str = "admin",
        timeout: int = 10,
        auth_mechanism: Optional[str] = "DEFAULT",
        ssl: Optional[bool] = True,
        logger: Optional[Any] = None,
        auth_retry_attempts: int = 3,
        auth_retry_wait: float = 1.0,
    ) -> None:
        # Import the asyncio client lazily to provide a clear error if unavailable
        AsyncMongoClient = None  # type: ignore
        import_error: Optional[Exception] = None
        try:
            from pymongo.asynchronous.mongo_client import AsyncMongoClient  # type: ignore
        except Exception as e1:  # pragma: no cover - fallback path
            import_error = e1
            try:
                # Some versions may expose it at package level
                from pymongo.asynchronous import AsyncMongoClient as _AltAsyncClient  # type: ignore
                AsyncMongoClient = _AltAsyncClient  # type: ignore
            except Exception as e2:  # pragma: no cover - fallback path
                import_error = e2
                try:
                    # Older preview namespace
                    from pymongo.asyncio import MongoClient as _AsyncioClient  # type: ignore
                    AsyncMongoClient = _AsyncioClient  # type: ignore
                except Exception as e3:  # pragma: no cover - final fallback
                    import_error = e3

        if AsyncMongoClient is None:  # type: ignore
            raise ImportError(
                "PyMongo async client is not available. Ensure a recent pymongo version is installed."
            ) from import_error

        self.client = AsyncMongoClient(  # type: ignore[misc]
            uri,
            username=username,
            password=password,
            authSource=auth_source,
            authMechanism=auth_mechanism,
            ssl=ssl,
            serverSelectionTimeoutMS=timeout * 1000,
        )
        self.logger = logger or setup_logger(__name__)
        self._log(
            f"Initialized AsyncMongoClient with authSource={auth_source}, "
            f"authMechanism={auth_mechanism}, ssl={ssl}"
        )
        self.auth_retry_attempts = auth_retry_attempts
        self.auth_retry_wait = auth_retry_wait

    # Async context manager
    async def __aenter__(self) -> "AsyncMongoConnector":
        await self._ping_with_retry()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    def _log(self, msg: str, level: str = "info") -> None:
        if self.logger:
            log_method = getattr(self.logger, level, self.logger.info)
            log_method(msg)

    async def _ping_with_retry(self) -> None:
        """Async ping to validate connection/auth, with retry."""
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(self.auth_retry_attempts),
            wait=wait_fixed(self.auth_retry_wait),
            reraise=True,
            retry=retry_if_exception_type(
                (OperationFailure, ServerSelectionTimeoutError, AutoReconnect, ConnectionFailure)
            ),
        ):
            with attempt:
                self._log("Pinging MongoDB (async) to verify connection/auth...", level="debug")
                await self.client.admin.command("ping")

    # Operations
    async def find(
        self,
        db_name: str,
        collection: str,
        filter: Dict[str, Any],
        projection: Optional[Dict[str, Any]] = None,
        batch_size: int = 1000,
    ) -> AsyncIterator[Dict[str, Any]]:
        """
        Async find documents with optional projection and paging.

        Args:
            db_name (str): Name of the database.
            collection (str): Name of the collection.
            filter (Dict[str, Any]): MongoDB filter document.
            projection (Optional[Dict[str, Any]]): Fields to include/exclude.
            batch_size (int): Number of documents per batch.

        Yields:
            Each document as a dictionary.
        """
        self._log(
            f"Executing async Mongo find on {db_name}.{collection} with filter: {filter}"
        )
        col = self.client[db_name][collection]
        cursor = col.find(filter, projection).batch_size(batch_size)
        async for doc in cursor:
            yield doc

    async def insert_many(
        self,
        db_name: str,
        collection: str,
        data: List[Dict[str, Any]],
        ordered: bool = False,
        batch_size: int = 1000,
    ) -> List[Any]:
        """
        Async insert multiple documents into a collection (batched).

        Note:
            PyMongo batches writes internally as well; manual batching is
            primarily for progress logging, memory control, and error isolation.
        """
        self._log(
            f"async insert_many: inserting {len(data)} docs into {db_name}.{collection} with batch_size={batch_size}"
        )
        col = self.client[db_name][collection]
        results: List[Any] = []
        for i in range(0, len(data), batch_size):
            batch = data[i : i + batch_size]
            result = await col.insert_many(batch, ordered=ordered)
            results.append(result)
        return results

    async def upsert_many(
        self,
        db_name: str,
        collection: str,
        data: List[Dict[str, Any]],
        unique_key: Optional[str],
        ordered: bool = False,
        batch_size: int = 1000,
    ) -> List[Any]:
        """
        Async upsert multiple documents using a unique key (batched).

        Uses `bulk_write` with `UpdateOne(..., upsert=True)` and `$set` to
        merge fields from each document. The `unique_key` must be present in
        each document and is used to build the filter.
        """
        if not unique_key:
            raise ValueError("unique_key must be provided for upsert_many")
        self._log(
            f"async upsert_many: upserting {len(data)} docs into {db_name}.{collection} with batch_size={batch_size}, unique_key={unique_key}"
        )
        col = self.client[db_name][collection]
        results: List[Any] = []
        for i in range(0, len(data), batch_size):
            batch = data[i : i + batch_size]
            operations = [
                UpdateOne({unique_key: doc[unique_key]}, {"$set": doc}, upsert=True)
                for doc in batch
                if unique_key in doc
            ]
            if operations:
                result = await col.bulk_write(operations, ordered=ordered)
                results.append(result)
        return results

    async def distinct(
        self,
        db_name: str,
        collection: str,
        key: str,
        filter: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Any]:
        """Async distinct values for a key, with optional filter."""
        self._log(
            f"Executing async Mongo distinct on {db_name}.{collection} for key='{key}' with filter: {filter or {}}"
        )
        col = self.client[db_name][collection]
        return await col.distinct(key, filter, **kwargs)

    async def delete(
        self,
        db_name: str,
        collection: str,
        filter: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        """Async delete a single document matching the filter."""
        self._log(
            f"Async deleting one from {db_name}.{collection} with filter: {filter}",
            level="info",
        )
        col = self.client[db_name][collection]
        return await col.delete_one(filter, **kwargs)

    async def delete_many(
        self,
        db_name: str,
        collection: str,
        filter: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        """Async delete all documents matching the filter."""
        self._log(
            f"Async deleting many from {db_name}.{collection} with filter: {filter}",
            level="info",
        )
        col = self.client[db_name][collection]
        return await col.delete_many(filter, **kwargs)

    async def close(self) -> None:
        """Close the underlying async client."""
        self._log("Closing AsyncMongoClient connection", level="debug")
        try:
            result = self.client.close()
            if inspect.isawaitable(result):
                await result
        except Exception as exc:  # pragma: no cover - best-effort close
            self._log(f"Error during AsyncMongoClient.close(): {exc}", level="warning")



import pyodbc
from typing import List, Dict, Generator, Any
from ppp_connectors.helpers import setup_logger


class ODBCConnector:
    """
    A connector class for interacting with ODBC-compatible databases.

    Provides methods for querying and bulk inserts.
    Supports use as a context manager for automatic connection cleanup.
    """

    def __init__(self, conn_str: str, logger: Any = None):
        """
        Initialize the ODBC connection.

        Args:
            conn_str (str): The ODBC connection string.
            logger (Any, optional): Logger instance for logging. Defaults to None.
        """
        self.conn = pyodbc.connect(conn_str)
        self.logger = logger or setup_logger(__name__)
        self._log("ODBC connection established")

    def _log(self, msg: str, level: str = "info"):
        if self.logger:
            log_method = getattr(self.logger, level, self.logger.info)
            log_method(msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """Close the ODBC connection."""
        if self.conn:
            self.conn.close()
            self._log("ODBC connection closed")

    def query(self, base_query: str) -> Generator[Dict[str, Any], None, None]:
        """
        Execute a query against an ODBC database and yield each row as a dictionary.

        Args:
            base_query (str): The SQL query to execute.

        Yields:
            Generator[Dict[str, Any], None, None]: A generator that yields each search hit as a dictionary.

        Logs:
            Execution of the query.

        Note:
            This method returns a generator. If you want to collect all results,
            you can wrap the result in `list()`, but beware of memory usage if the
            result set is large.
        """
        self._log(f"Executing ODBC query")
        cursor = self.conn.cursor()
        cursor.execute(base_query)
        columns = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            yield dict(zip(columns, row))

    def bulk_insert(self, table: str, data: List[Dict[str, Any]]):
        """
        Perform a bulk insert into an ODBC database table.

        Args:
            table (str): Name of the table to insert into.
            data (List[Dict[str, Any]]): List of rows to insert.

        Returns:
            None

        Logs:
            Number of rows inserted and target table.
        """
        if not data:
            return
        self._log(f"Inserting {len(data)} rows into table {table}")
        columns = list(data[0].keys())
        placeholders = ", ".join(["?"] * len(columns))
        insert_query = f"INSERT INTO {table} ({', '.join(columns)}) VALUES ({placeholders})"
        values = [tuple(row[col] for col in columns) for row in data]
        cursor = self.conn.cursor()
        cursor.fast_executemany = True
        cursor.executemany(insert_query, values)
        self.conn.commit()



import pytest
import httpx
from unittest.mock import AsyncMock, patch
from ppp_connectors.api_connectors.generic import AsyncGenericConnector


@pytest.mark.asyncio
async def test_async_init_sets_base_url():
    connector = AsyncGenericConnector(base_url="https://example.com")
    assert connector.base_url == "https://example.com"


@patch("ppp_connectors.api_connectors.generic.AsyncGenericConnector._make_request", new_callable=AsyncMock)
@pytest.mark.asyncio
async def test_async_get_request(mock_make_request):
    mock_response = httpx.Response(200, json={"result": "ok"})
    mock_make_request.return_value = mock_response

    connector = AsyncGenericConnector(base_url="https://example.com")
    response = await connector.request("GET", "/test")

    assert response.status_code == 200
    assert response.json() == {"result": "ok"}
    mock_make_request.assert_awaited_once_with(
        method="GET",
        endpoint="/test",
        headers=connector.headers,
        params=None,
        json=None,
        auth=None,
        retry_kwargs=None,
    )


@patch("ppp_connectors.api_connectors.generic.AsyncGenericConnector._make_request", new_callable=AsyncMock)
@pytest.mark.asyncio
async def test_async_post_request(mock_make_request):
    mock_response = httpx.Response(200, json={"posted": True})
    mock_make_request.return_value = mock_response

    connector = AsyncGenericConnector(base_url="https://example.com")
    response = await connector.request("POST", "/submit", json={"key": "value"})

    assert response.status_code == 200
    assert response.json() == {"posted": True}
    mock_make_request.assert_awaited_once_with(
        method="POST",
        endpoint="/submit",
        headers=connector.headers,
        params=None,
        json={"key": "value"},
        auth=None,
        retry_kwargs=None,
    )


import httpx
from typing import Dict, Any, Optional
from ppp_connectors.api_connectors.broker import AsyncBroker, Broker, bubble_broker_init_signature, log_method_call


@bubble_broker_init_signature()
class SpycloudConnector(Broker):
    """
    SpyCloudConnector provides typed methods to interact with various SpyCloud APIs, including:
    - SIP Cookie Domains
    - ATO Breach Catalog
    - ATO Search
    - Investigations Search
    """

    def __init__(self, sip_key: Optional[str] = None, ato_key: Optional[str] = None,
                 inv_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://api.spycloud.io", **kwargs)
        self.sip_key = sip_key or self.env_config.get("SPYCLOUD_API_SIP_KEY")
        self.ato_key = ato_key or self.env_config.get("SPYCLOUD_API_ATO_KEY")
        self.inv_key = inv_key or self.env_config.get("SPYCLOUD_API_INV_KEY")

    @log_method_call
    def sip_cookie_domains(self, cookie_domains: str, **kwargs) -> httpx.Response:
        """Query SIP cookie domain data."""
        if not self.sip_key:
            raise ValueError("SPYCLOUD_API_SIP_KEY is required for this request.")
        endpoint = f"/sip-v1/breach/data/cookie-domains/{cookie_domains}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.sip_key
        }
        return self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)

    @log_method_call
    def ato_breach_catalog(self, query: str, **kwargs) -> httpx.Response:
        """Query ATO breach catalog."""
        if not self.ato_key:
            raise ValueError("SPYCLOUD_API_ATO_KEY is required for this request.")
        endpoint = "/sp-v2/breach/catalog"
        headers = {
            "accept": "application/json",
            "x-api-key": self.ato_key
        }
        params = {"query": query, **kwargs}
        return self._make_request("get", endpoint=endpoint, headers=headers, params=params)

    @log_method_call
    def ato_search(self, search_type: str, query: str, **kwargs) -> httpx.Response:
        """Search against SpyCloud's ATO breach dataset."""
        if not self.ato_key:
            raise ValueError("SPYCLOUD_API_ATO_KEY is required for this request.")

        base_url = "/sp-v2/breach/data"
        endpoints = {
            'domain': 'domains',
            'email': 'emails',
            'ip': 'ips',
            'username': 'usernames',
            'phone-number': 'phone-numbers',
        }

        if search_type not in endpoints:
            raise ValueError(f'Invalid search_type: {search_type}. Must be one of: {", ".join(endpoints.keys())}')

        endpoint = f"{base_url}/{endpoints[search_type]}/{query}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.ato_key
        }
        return self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)

    @log_method_call
    def investigations_search(self, search_type: str, query: str, **kwargs) -> httpx.Response:
        """Search SpyCloud Investigations API by type and query."""
        if not self.inv_key:
            raise ValueError("SPYCLOUD_API_INV_KEY is required for this request.")

        base_url = "/investigations-v2/breach/data"
        endpoints = {
            'domain': 'domains',
            'email': 'emails',
            'ip': 'ips',
            'infected-machine-id': 'infected-machine-ids',
            'log-id': 'log-ids',
            'password': 'passwords',
            'username': 'usernames',
            'email-username': 'email-usernames',
            'phone-number': 'phone-numbers',
            'social-handle': 'social-handles',
            'bank-number': 'bank-numbers',
            'cc-number': 'cc-numbers',
            'drivers-license': 'drivers-licenses',
            'national-id': 'national-ids',
            'passport-number': 'passport-numbers',
            'ssn': 'social-security-numbers',
        }

        if search_type not in endpoints:
            raise ValueError(f'Invalid search_type: {search_type}. Must be one of: {", ".join(endpoints.keys())}')

        endpoint = f"{base_url}/{endpoints[search_type]}/{query}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.inv_key
        }
        return self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)


@bubble_broker_init_signature()
class AsyncSpycloudConnector(AsyncBroker):
    """
    Async version of SpycloudConnector.
    """

    def __init__(self, sip_key: Optional[str] = None, ato_key: Optional[str] = None,
                 inv_key: Optional[str] = None, **kwargs):
        super().__init__(base_url="https://api.spycloud.io", **kwargs)
        self.sip_key = sip_key or self.env_config.get("SPYCLOUD_API_SIP_KEY")
        self.ato_key = ato_key or self.env_config.get("SPYCLOUD_API_ATO_KEY")
        self.inv_key = inv_key or self.env_config.get("SPYCLOUD_API_INV_KEY")

    @log_method_call
    async def sip_cookie_domains(self, cookie_domains: str, **kwargs) -> httpx.Response:
        """Query SIP cookie domain data (async)."""
        if not self.sip_key:
            raise ValueError("SPYCLOUD_API_SIP_KEY is required for this request.")
        endpoint = f"/sip-v1/breach/data/cookie-domains/{cookie_domains}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.sip_key
        }
        return await self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)

    @log_method_call
    async def ato_breach_catalog(self, query: str, **kwargs) -> httpx.Response:
        """Query ATO breach catalog (async)."""
        if not self.ato_key:
            raise ValueError("SPYCLOUD_API_ATO_KEY is required for this request.")
        endpoint = "/sp-v2/breach/catalog"
        headers = {
            "accept": "application/json",
            "x-api-key": self.ato_key
        }
        params = {"query": query, **kwargs}
        return await self._make_request("get", endpoint=endpoint, headers=headers, params=params)

    @log_method_call
    async def ato_search(self, search_type: str, query: str, **kwargs) -> httpx.Response:
        """Search against SpyCloud's ATO breach dataset (async)."""
        if not self.ato_key:
            raise ValueError("SPYCLOUD_API_ATO_KEY is required for this request.")

        base_url = "/sp-v2/breach/data"
        endpoints = {
            'domain': 'domains',
            'email': 'emails',
            'ip': 'ips',
            'username': 'usernames',
            'phone-number': 'phone-numbers',
        }

        if search_type not in endpoints:
            raise ValueError(f'Invalid search_type: {search_type}. Must be one of: {", ".join(endpoints.keys())}')

        endpoint = f"{base_url}/{endpoints[search_type]}/{query}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.ato_key
        }
        return await self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)

    @log_method_call
    async def investigations_search(self, search_type: str, query: str, **kwargs) -> httpx.Response:
        """Search SpyCloud Investigations API by type and query (async)."""
        if not self.inv_key:
            raise ValueError("SPYCLOUD_API_INV_KEY is required for this request.")

        base_url = "/investigations-v2/breach/data"
        endpoints = {
            'domain': 'domains',
            'email': 'emails',
            'ip': 'ips',
            'infected-machine-id': 'infected-machine-ids',
            'log-id': 'log-ids',
            'password': 'passwords',
            'username': 'usernames',
            'email-username': 'email-usernames',
            'phone-number': 'phone-numbers',
            'social-handle': 'social-handles',
            'bank-number': 'bank-numbers',
            'cc-number': 'cc-numbers',
            'drivers-license': 'drivers-licenses',
            'national-id': 'national-ids',
            'passport-number': 'passport-numbers',
            'ssn': 'social-security-numbers',
        }

        if search_type not in endpoints:
            raise ValueError(f'Invalid search_type: {search_type}. Must be one of: {", ".join(endpoints.keys())}')

        endpoint = f"{base_url}/{endpoints[search_type]}/{query}"
        headers = {
            "accept": "application/json",
            "x-api-key": self.inv_key
        }
        return await self._make_request("get", endpoint=endpoint, headers=headers, params=kwargs)


import httpx
import pytest
from ppp_connectors.api_connectors.urlscan import URLScanConnector

@pytest.mark.integration
def test_urlscan_results_vcr(vcr_cassette):
    with vcr_cassette.use_cassette("test_urlscan_results_vcr"):
        connector = URLScanConnector(load_env_vars=True, enable_logging=True)
        result = connector.results("01958568-c986-7001-816d-9e0ccd7c4c4a")

        assert isinstance(result, httpx.Response)
        assert "task" in result.json()


import httpx
from typing import Generator, Dict, Any, Optional
from ppp_connectors.helpers import setup_logger


class SplunkConnector:
    """
    A connector class for interacting with Splunk via its REST API.

    Provides methods for submitting search jobs and streaming paginated results.
    """
    def __init__(
        self,
        host: str,
        port: int,
        username: Optional[str] = None,
        password: Optional[str] = None,
        scheme: str = "https",
        verify: bool = True,
        timeout: int = 30,
        logger: Optional[Any] = None
    ):
        """
        Initialize the SplunkConnector with connection details.

        Args:
            host (str): The Splunk server host.
            port (int): The Splunk management port.
            username (Optional[str]): Username for authentication. Defaults to None.
            password (Optional[str]): Password for authentication. Defaults to None.
            scheme (str): HTTP or HTTPS. Defaults to "https".
            verify (bool): Whether to verify SSL certificates. Defaults to True.
            timeout (int): Request timeout in seconds. Defaults to 30.
            logger (Optional[Any]): Logger instance. If provided, actions will be logged.
        """
        self.base_url = f"{scheme}://{host}:{port}"
        self.auth = (username, password)
        self.verify = verify
        self.timeout = timeout
        self.logger = logger or setup_logger(__name__)

        # Suppress InsecureRequestWarnings if the user sets verify=False
        if not self.verify:
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _log(self, msg: str, level: str = "info"):
        if self.logger:
            log_method = getattr(self.logger, level, self.logger.info)
            log_method(msg)

    def query(
        self,
        search: str,
        count: int = 1000,
        earliest_time: Optional[str] = None,
        latest_time: Optional[str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Submit a search job to Splunk and stream results as dictionaries.

        Args:
            search (str): The search query string.
            count (int): Number of results per batch. Defaults to 1000.
            earliest_time (Optional[str]): Earliest time for the search. Defaults to None.
            latest_time (Optional[str]): Latest time for the search. Defaults to None.

        Yields:
            Dict[str, Any]: Each search result as a dictionary.

        Logs actions if logger is enabled.
        """
        # 1️⃣ Create job
        self._log(f"Submitting search job: {search}")
        data = {
            "search": search,
            "output_mode": "json",
            "count": count
        }
        if earliest_time:
            data["earliest_time"] = earliest_time
        if latest_time:
            data["latest_time"] = latest_time

        create_resp = httpx.post(
            f"{self.base_url}/services/search/jobs",
            auth=self.auth,
            data=data,
            verify=self.verify,
            timeout=self.timeout
        )
        create_resp.raise_for_status()
        sid = create_resp.json()["sid"]

        # 2️⃣ Poll until ready
        while True:
            self._log(f"Polling job {sid} status...")
            status_resp = httpx.get(
                f"{self.base_url}/services/search/jobs/{sid}",
                auth=self.auth,
                params={"output_mode": "json"},
                verify=self.verify,
                timeout=self.timeout
            )
            status_resp.raise_for_status()
            content = status_resp.json()
            if content["entry"][0]["content"]["isDone"]:
                break

        # 3️⃣ Fetch results
        offset = 0
        while True:
            self._log(f"Fetching results batch starting at offset {offset}")
            results_resp = httpx.get(
                f"{self.base_url}/services/search/jobs/{sid}/results",
                auth=self.auth,
                params={
                    "output_mode": "json",
                    "count": count,
                    "offset": offset
                },
                verify=self.verify,
                timeout=self.timeout
            )
            results_resp.raise_for_status()
            results = results_resp.json().get("results", [])
            if not results:
                break
            for row in results:
                yield row
            offset += len(results)



# API connectors
from ppp_connectors.api_connectors import (
    urlscan,
    spycloud,
    twilio,
    flashpoint,
    ipqs,
    generic,
)

# DBMS connectors
from ppp_connectors.dbms_connectors import (
    elasticsearch,
    mongo,
    odbc,
    splunk
)

# Export the modules and re-exports
__all__ = [
    "elasticsearch",
    "flashpoint",
    "generic",
    "ipqs",
    "mongo",
    "odbc",
    "splunk",
    "spycloud",
    "twilio",
    "urlscan",
]


import sys
import types
import pytest
from unittest.mock import AsyncMock

from ppp_connectors.dbms_connectors.mongo_async import AsyncMongoConnector


@pytest.mark.asyncio
async def test_async_init_and_ping(monkeypatch):
    fake_client = types.SimpleNamespace()
    fake_client.admin = types.SimpleNamespace()
    fake_client.admin.command = AsyncMock(return_value={"ok": 1})

    # Provide pymongo.asyncio.MongoClient for import inside the connector
    ns = types.SimpleNamespace(AsyncMongoClient=lambda *a, **k: fake_client)
    monkeypatch.setitem(sys.modules, "pymongo.asynchronous.mongo_client", ns)

    async with AsyncMongoConnector(uri="mongodb://localhost:27017") as conn:
        assert isinstance(conn, AsyncMongoConnector)

    fake_client.admin.command.assert_awaited_once_with("ping")


class _AsyncCursor:
    def __init__(self, items):
        self._items = items
        self._idx = 0

    def batch_size(self, n):
        return self

    def __aiter__(self):
        self._idx = 0
        return self

    async def __anext__(self):
        if self._idx >= len(self._items):
            raise StopAsyncIteration
        item = self._items[self._idx]
        self._idx += 1
        return item


@pytest.mark.asyncio
async def test_async_find(monkeypatch):
    docs = [{"_id": 1}, {"_id": 2}]

    class _FakeCollection:
        def find(self, f, p):
            return _AsyncCursor(docs)

    class _FakeDB(dict):
        def __getitem__(self, k):
            return _FakeCollection()

    class _FakeClient:
        def __init__(self):
            self.admin = types.SimpleNamespace()
            self.admin.command = AsyncMock(return_value={"ok": 1})

        def __getitem__(self, k):
            return _FakeDB()

        def close(self):
            pass

    fake_client = _FakeClient()
    ns = types.SimpleNamespace(AsyncMongoClient=lambda *a, **k: fake_client)
    monkeypatch.setitem(sys.modules, "pymongo.asynchronous.mongo_client", ns)

    async with AsyncMongoConnector(uri="mongodb://localhost:27017") as conn:
        out = [doc async for doc in conn.find("db", "col", filter={})]

    assert out == docs



import httpx
import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.flashpoint import FlashpointConnector

def test_init_with_api_key():
    connector = FlashpointConnector(api_key="test_token")
    assert connector.api_key == "test_token"
    assert connector.headers["Authorization"] == "Bearer test_token"


@patch.dict("os.environ", {"FLASHPOINT_API_KEY": "env_token"}, clear=True)
def test_init_with_env_key():
    connector = FlashpointConnector(load_env_vars=True)
    assert connector.api_key == "env_token"
    assert connector.headers["Authorization"] == "Bearer env_token"


def test_init_missing_key():
    with patch.dict("os.environ", {}, clear=True):
        with pytest.raises(ValueError, match="FLASHPOINT_API_KEY is required"):
            FlashpointConnector()


@patch("ppp_connectors.api_connectors.flashpoint.FlashpointConnector.post")
def test_search_fraud(mock_post):
    import json

    # Use a real httpx.Response so our test matches the new return type
    request = httpx.Request("POST", "https://api.flashpoint.io/mock")
    payload = {"success": True, "data": []}
    mock_response = httpx.Response(
        200,
        request=request,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_post.return_value = mock_response

    connector = FlashpointConnector(api_key="mock_token")
    result = connector.search_fraud("credential stuffing")

    assert isinstance(result, httpx.Response)
    assert result.json() == payload
    mock_post.assert_called_once()


import httpx
import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.urlscan import URLScanConnector

def test_init_with_api_key():
    connector = URLScanConnector(api_key="test_key")
    assert connector.api_key == "test_key"

@patch.dict("os.environ", {"URLSCAN_API_KEY": "env_key"}, clear=True)
def test_init_with_env_key():
    connector = URLScanConnector(load_env_vars=True)
    assert connector.api_key == "env_key"

@patch("ppp_connectors.api_connectors.broker.combine_env_configs", return_value={})
def test_init_missing_auth_keys(mock_env):
    with pytest.raises(ValueError, match="API key is required for URLScanConnector"):
        URLScanConnector(load_env_vars=True)

@patch("ppp_connectors.api_connectors.urlscan.URLScanConnector.get")
def test_results(mock_get):
    import json

    # Build a real httpx.Response to match the new return type
    request = httpx.Request("GET", "https://urlscan.io/api/v1/result/abc123")
    payload = {"task": "test"}
    mock_response = httpx.Response(
        200,
        request=request,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_get.return_value = mock_response

    connector = URLScanConnector(api_key="test_key")
    result = connector.results("abc123")

    assert isinstance(result, httpx.Response)
    assert result.json() == payload


from datetime import datetime
from dotenv import dotenv_values, find_dotenv
import logging
import os
import sys
from typing import Dict, Set, List, Any, Optional


def check_required_env_vars(config: Dict[str, str], required_vars: List[str]) -> None:
    """Ensure that the env variables required for a function are present either in \
        the .env file, or in the system's environment variables.

    Args:
        config (Dict[str, str]): the env_config variable that contains values from the .env file
        required_vars (List[str]): the env variables required for a function to successfully function
    """

    dotenv_missing_vars: Set[str] = set(required_vars) - set(config.keys())
    osenv_missing_vars: Set[str] = set(required_vars) - set(os.environ)
    missing_vars = dotenv_missing_vars | osenv_missing_vars

    if dotenv_missing_vars and osenv_missing_vars:
        print(f'[!] Error: missing required environment variables: {", ".join(missing_vars)}. '
              'Please ensure these are present either in your .env file, or in the '
              'system\'s environment variables.', file=sys.stderr)
        sys.exit(1)


def combine_env_configs() -> Dict[str, Any]:
    """Find a .env file if it exists, and combine it with system environment
        variables to form a "combined_config" dictionary of environment variables

    Returns:
        Dict: a dictionary containing the output of a .env file (if found), and
        system environment variables
    """

    env_config: Dict[str, Any] = dict(dotenv_values(find_dotenv()))

    combined_config: Dict[str, Any] = {**env_config, **dict(os.environ)}

    return combined_config


def validate_date_string(date_str: str) -> bool:
    """Validates that a date string is, well, a valid date string

    Args:
        date_str (str): a string in "YYYY-MM-DD" format

    Returns:
        bool: True or False as valid or not
    """
    try:
        datetime.strptime(date_str, "%Y-%m-%d")
        return True
    except ValueError:
        return False


def setup_logger(
    name: str = __name__,
    level: int = logging.INFO,
    log_file: Optional[str] = None,
    use_stdout: bool = True
) -> logging.Logger:
    """
    Configures and returns a logger with optional StreamHandler and/or FileHandler.

    This function checks if a logger with the specified name already has any StreamHandler
    or FileHandler attached, and if not, it adds them according to the parameters.

    Args:
        name (str): The name of the logger to configure.
        level (int): The logging level to set for the logger. Defaults to logging.INFO.
        log_file (Optional[str]): If provided, logs will be written to this file.
        use_stdout (bool): Whether to log to standard output. Defaults to True.

    Returns:
        logging.Logger: A configured logger instance.
    """
    logger: logging.Logger = logging.getLogger(name)

    logger.setLevel(level)

    formatter = logging.Formatter(
        '[%(asctime)s]\t%(levelname)s\t%(name)s:%(lineno)d\t%(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    if use_stdout and not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    if log_file and not any(isinstance(h, logging.FileHandler) for h in logger.handlers):
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger.propagate = False
    return logger



import pytest
import httpx
from ppp_connectors.api_connectors.spycloud import AsyncSpycloudConnector

@pytest.mark.asyncio
async def test_async_init_with_env_keys(monkeypatch):
    monkeypatch.setenv("SPYCLOUD_API_SIP_KEY", "sip")
    monkeypatch.setenv("SPYCLOUD_API_ATO_KEY", "ato")
    monkeypatch.setenv("SPYCLOUD_API_INV_KEY", "inv")

    conn = AsyncSpycloudConnector(load_env_vars=True)
    assert conn.sip_key == "sip"
    assert conn.ato_key == "ato"
    assert conn.inv_key == "inv"

@pytest.mark.asyncio
async def test_async_sip_cookie_domains_sends(mocker):
    mock_response = mocker.AsyncMock(spec=httpx.Response)
    request = mocker.patch("httpx.AsyncClient.request", return_value=mock_response)

    connector = AsyncSpycloudConnector(sip_key="abc123")
    result = await connector.sip_cookie_domains("test.com", q="xyz")

    assert result is mock_response
    request.assert_called_once()
    args, kwargs = request.call_args
    assert kwargs["url"].endswith("/sip-v1/breach/data/cookie-domains/test.com")
    assert kwargs["headers"]["x-api-key"] == "abc123"
    assert kwargs["params"]["q"] == "xyz"

@pytest.mark.asyncio
async def test_async_ato_search_ip(mocker):
    mock_response = mocker.AsyncMock(spec=httpx.Response)
    mocker.patch("httpx.AsyncClient.request", return_value=mock_response)

    connector = AsyncSpycloudConnector(ato_key="abc123")
    resp = await connector.ato_search("ip", "1.2.3.4")
    assert resp is mock_response

@pytest.mark.asyncio
async def test_async_investigations_invalid_type_raises():
    conn = AsyncSpycloudConnector(inv_key="abc")
    with pytest.raises(ValueError):
        await conn.investigations_search("not-a-real-type", "foo")


import httpx
import pytest
from unittest.mock import patch, MagicMock
from ppp_connectors.api_connectors.ipqs import IPQSConnector


def test_init_with_api_key():
    connector = IPQSConnector(api_key="test_key")
    assert connector.api_key == "test_key"
    assert connector.headers["Content-Type"] == "application/json"


def test_init_with_env_key():
    with patch.dict("os.environ", {"IPQS_API_KEY": "env_key"}):
        connector = IPQSConnector(load_env_vars=True)
        assert connector.api_key == "env_key"


def test_init_missing_key():
    with patch.dict("os.environ", {}, clear=True):
        with pytest.raises(ValueError, match="API key is required"):
            IPQSConnector()


@patch("ppp_connectors.api_connectors.ipqs.IPQSConnector.post")
def test_malicious_url(mock_post):
    # Build a real httpx.Response to match the new return type
    import json

    request = httpx.Request("POST", "https://ipqualityscore.com/api/json/url/")
    payload = {"success": True, "domain": "example.com"}
    mock_response = httpx.Response(
        200,
        request=request,
        content=json.dumps(payload).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
    mock_post.return_value = mock_response

    connector = IPQSConnector(api_key="test_key")
    result = connector.malicious_url("example.com")

    mock_post.assert_called_once()
    assert isinstance(result, httpx.Response)
    assert result.json() == payload
